# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import (
    TYPE_CHECKING,
    Iterable,
    List,
    cast,
    Optional,
    Callable,
    Dict,
    Union,
    Type,
)
import logging
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000, DXF2010
from ezdxf.math import Vec3, Vec2, BoundingBox2d
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .dxfobj import DXFObject
from .factory import register_entity

logger = logging.getLogger("ezdxf")

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter,
        DXFNamespace,
        Drawing,
        Vertex,
        DXFTag,
        Matrix44,
        Auditor,
        DXFEntity,
    )

__all__ = ["Image", "ImageDef", "ImageDefReactor", "RasterVariables", "Wipeout"]


class ImageBase(DXFGraphic):
    """DXF IMAGE entity"""

    DXFTYPE = "IMAGEBASE"
    _CLS_GROUP_CODES: Dict[int, Union[str, List[str]]] = dict()
    _SUBCLASS_NAME = "dummy"
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    SHOW_IMAGE = 1
    SHOW_IMAGE_WHEN_NOT_ALIGNED = 2
    USE_CLIPPING_BOUNDARY = 4
    USE_TRANSPARENCY = 8

    def __init__(self):
        super().__init__()
        # Boundary/Clipping path coordinates:
        # 0/0 is in the Left/Top corner of the image!
        # x-coordinates increases in u_pixel vector direction
        # y-coordinates increases against the v_pixel vector!
        # see also WCS coordinate calculation
        self._boundary_path: List[Vec2] = []

    def _copy_data(self, entity: "DXFEntity") -> None:
        assert isinstance(entity, ImageBase)
        entity._boundary_path = list(self._boundary_path)

    def post_new_hook(self) -> None:
        super().post_new_hook()
        self.reset_boundary_path()

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            path_tags = processor.subclasses[2].pop_tags(codes=(14,))
            self.load_boundary_path(path_tags)
            processor.fast_load_dxfattribs(
                dxf, self._CLS_GROUP_CODES, 2, recover=True
            )
            if len(self.boundary_path) < 2:  # something is wrong
                self.dxf = dxf
                self.reset_boundary_path()
        return dxf

    def load_boundary_path(self, tags: Iterable["DXFTag"]):
        self._boundary_path = [
            Vec2(value) for code, value in tags if code == 14
        ]

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, self._SUBCLASS_NAME)
        self.dxf.count_boundary_points = len(self.boundary_path)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "class_version",
                "insert",
                "u_pixel",
                "v_pixel",
                "image_size",
                "image_def_handle",
                "flags",
                "clipping",
                "brightness",
                "contrast",
                "fade",
                "image_def_reactor_handle",
                "clipping_boundary_type",
                "count_boundary_points",
            ],
        )
        self.export_boundary_path(tagwriter)
        if tagwriter.dxfversion >= DXF2010:
            self.dxf.export_dxf_attribs(tagwriter, "clip_mode")

    def export_boundary_path(self, tagwriter: "TagWriter"):
        for vertex in self.boundary_path:
            tagwriter.write_vertex(14, vertex)

    @property
    def boundary_path(self):
        """A list of vertices as pixel coordinates, Two vertices describe a
        rectangle, lower left corner is ``(-0.5, -0.5)`` and upper right corner
        is ``(ImageSizeX-0.5, ImageSizeY-0.5)``, more than two vertices is a
        polygon as clipping path. All vertices as pixel coordinates. (read/write)

        """
        return self._boundary_path

    @boundary_path.setter
    def boundary_path(self, vertices: Iterable["Vertex"]) -> None:
        self.set_boundary_path(vertices)

    def set_boundary_path(self, vertices: Iterable["Vertex"]) -> None:
        """Set boundary path to `vertices`. Two vertices describe a rectangle
        (lower left and upper right corner), more than two vertices is a polygon
        as clipping path.

        """
        _vertices = Vec2.list(vertices)
        if len(_vertices):
            if len(_vertices) > 2 and not _vertices[-1].isclose(_vertices[0]):
                # Close path, otherwise AutoCAD crashes
                _vertices.append(_vertices[0])
            self._boundary_path = _vertices
            self.set_flag_state(self.USE_CLIPPING_BOUNDARY, state=True)
            self.dxf.clipping = 1
            self.dxf.clipping_boundary_type = 1 if len(_vertices) < 3 else 2
            self.dxf.count_boundary_points = len(self._boundary_path)
        else:
            self.reset_boundary_path()

    def reset_boundary_path(self) -> None:
        """Reset boundary path to the default rectangle [(-0.5, -0.5),
        (ImageSizeX-0.5, ImageSizeY-0.5)].

        """
        lower_left_corner = Vec2(-0.5, -0.5)
        upper_right_corner = Vec2(self.dxf.image_size) + lower_left_corner
        self._boundary_path = [lower_left_corner, upper_right_corner]
        self.set_flag_state(Image.USE_CLIPPING_BOUNDARY, state=False)
        self.dxf.clipping = 0
        self.dxf.clipping_boundary_type = 1
        self.dxf.count_boundary_points = 2

    def transform(self, m: "Matrix44") -> "ImageBase":
        """Transform IMAGE entity by transformation matrix `m` inplace."""
        self.dxf.insert = m.transform(self.dxf.insert)
        self.dxf.u_pixel = m.transform_direction(self.dxf.u_pixel)
        self.dxf.v_pixel = m.transform_direction(self.dxf.v_pixel)
        self.post_transform(m)
        return self

    def boundary_path_wcs(self) -> List[Vec3]:
        """Returns the boundary/clipping path in WCS coordinates.

        .. versionadded:: 0.14

        Since version 0.16 it's recommended to create the clipping path
        as :class:`~ezdxf.path.Path` object by the
        :func:`~ezdxf.path.make_path` function::

            form ezdxf.path import make_path

            image = ...  # get image entity
            clipping_path = make_path(image)

        """

        u = Vec3(self.dxf.u_pixel)
        v = Vec3(self.dxf.v_pixel)
        origin = Vec3(self.dxf.insert)
        origin += u * 0.5 - v * 0.5
        height = self.dxf.image_size.y
        boundary_path = self.boundary_path
        if len(boundary_path) == 2:  # rectangle
            p0, p1 = boundary_path
            boundary_path = [p0, Vec2(p1.x, p0.y), p1, Vec2(p0.x, p1.y)]
        # Boundary/Clipping path origin 0/0 is in the Left/Top corner
        # of the image!
        vertices = [
            origin + (u * p.x) + (v * (height - p.y)) for p in boundary_path
        ]
        if not vertices[0].isclose(vertices[-1]):
            vertices.append(vertices[0])
        return vertices

    def destroy(self) -> None:
        if not self.is_alive:
            return

        del self._boundary_path
        super().destroy()


acdb_image = DefSubclass(
    "AcDbRasterImage",
    {
        "class_version": DXFAttr(90, dxfversion=DXF2000, default=0),
        "insert": DXFAttr(10, xtype=XType.point3d),
        # U-vector of a single pixel (points along the visual bottom of the image,
        # starting at the insertion point)
        "u_pixel": DXFAttr(11, xtype=XType.point3d),
        # V-vector of a single pixel (points along the visual left side of the
        # image, starting at the insertion point)
        "v_pixel": DXFAttr(12, xtype=XType.point3d),
        # Image size in pixels
        "image_size": DXFAttr(13, xtype=XType.point2d),
        # Hard reference to image def object
        "image_def_handle": DXFAttr(340),
        # Image display properties:
        # 1 = Show image
        # 2 = Show image when not aligned with screen
        # 4 = Use clipping boundary
        # 8 = Transparency is on
        "flags": DXFAttr(70, default=3),
        # Clipping state:
        # 0 = Off
        # 1 = On
        "clipping": DXFAttr(
            280,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Brightness value (0-100; default = 50)
        "brightness": DXFAttr(
            281,
            default=50,
            validator=validator.is_in_integer_range(0, 101),
            fixer=validator.fit_into_integer_range(0, 101),
        ),
        # Contrast value (0-100; default = 50)
        "contrast": DXFAttr(
            282,
            default=50,
            validator=validator.is_in_integer_range(0, 101),
            fixer=validator.fit_into_integer_range(0, 101),
        ),
        # Fade value (0-100; default = 0)
        "fade": DXFAttr(
            283,
            default=0,
            validator=validator.is_in_integer_range(0, 101),
            fixer=validator.fit_into_integer_range(0, 101),
        ),
        # Hard reference to image def reactor object, not required by AutoCAD
        "image_def_reactor_handle": DXFAttr(360),
        # Clipping boundary type:
        # 1 = Rectangular
        # 2 = Polygonal
        "clipping_boundary_type": DXFAttr(
            71,
            default=1,
            validator=validator.is_one_of({1, 2}),
            fixer=RETURN_DEFAULT,
        ),
        # Number of clip boundary vertices that follow
        "count_boundary_points": DXFAttr(91),
        # Clip mode:
        # 0 = outside
        # 1 = inside mode
        "clip_mode": DXFAttr(
            290,
            dxfversion=DXF2010,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # boundary path coordinates are pixel coordinates NOT drawing units
    },
)
acdb_image_group_codes = group_code_mapping(acdb_image)


@register_entity
class Image(ImageBase):
    """DXF IMAGE entity"""

    DXFTYPE = "IMAGE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_image)
    _CLS_GROUP_CODES = acdb_image_group_codes
    _SUBCLASS_NAME = acdb_image.name  # type: ignore
    DEFAULT_ATTRIBS = {"layer": "0", "flags": 3}

    def __init__(self):
        super().__init__()
        self._boundary_path: List[Vec2] = []
        self._image_def: Optional[ImageDef] = None
        self._image_def_reactor: Optional[ImageDefReactor] = None

    @classmethod
    def new(
        cls: Type["Image"],
        handle: str = None,
        owner: str = None,
        dxfattribs: Dict = None,
        doc: "Drawing" = None,
    ) -> "Image":
        dxfattribs = dxfattribs or {}
        # 'image_def' is not a real DXF attribute (image_def_handle)
        image_def = dxfattribs.pop("image_def", None)
        image_size = (1, 1)
        if image_def and image_def.is_alive:
            image_size = image_def.dxf.image_size
        dxfattribs.setdefault("image_size", image_size)

        image = cast("Image", super().new(handle, owner, dxfattribs, doc))
        image.image_def = image_def
        return image

    def copy(self) -> "Image":
        image_copy = cast("Image", super().copy())
        # Each Image has its own ImageDefReactor object,
        # which will be created by binding the copy to the
        # document.
        image_copy.dxf.discard("image_def_reactor_handle")
        image_copy._image_def_reactor = None
        image_copy._image_def = self._image_def
        return image_copy

    def post_bind_hook(self) -> None:
        # Document in LOAD process -> post_load_hook()
        if self.doc.is_loading:  # type: ignore
            return
        if self._image_def_reactor:  # ImageDefReactor already exist
            return
        # The new Image was created by ezdxf and the ImageDefReactor
        # object does not exist:
        self._create_image_def_reactor()

    def post_load_hook(self, doc: "Drawing") -> Optional[Callable]:
        super().post_load_hook(doc)
        db = doc.entitydb
        self._image_def = db.get(self.dxf.get("image_def_handle", None))
        if self._image_def is None:
            # unrecoverable structure error
            self.destroy()
            return None

        self._image_def_reactor = db.get(
            self.dxf.get("image_def_reactor_handle", None)
        )
        if self._image_def_reactor is None:
            # Image and ImageDef exist - this is recoverable by creating
            # an ImageDefReactor, but the objects section does not exist yet!
            # Return a post init command:
            return self._fix_missing_image_def_reactor
        return None

    def _fix_missing_image_def_reactor(self):
        try:
            self._create_image_def_reactor()
        except Exception as e:
            logger.exception(
                f"An exception occurred while executing fixing command for "
                f"{str(self)}, destroying entity.",
                exc_info=e,
            )
            self.destroy()
            return
        logger.debug(f"Created missing ImageDefReactor for {str(self)}")

    def _create_image_def_reactor(self):
        # ImageDef -> ImageDefReactor -> Image
        image_def_reactor = self.doc.objects.add_image_def_reactor(
            self.dxf.handle
        )
        reactor_handle = image_def_reactor.dxf.handle
        # Link Image to ImageDefReactor:
        self.dxf.image_def_reactor_handle = reactor_handle
        self._image_def_reactor = image_def_reactor
        # Link ImageDef to ImageDefReactor:
        self._image_def.append_reactor_handle(reactor_handle)

    @property
    def image_def(self) -> Optional["ImageDef"]:
        """Returns the associated IMAGEDEF entity, see :class:`ImageDef`."""
        return self._image_def

    @image_def.setter
    def image_def(self, image_def: "ImageDef") -> None:
        if image_def and image_def.is_alive:
            self.dxf.image_def_handle = image_def.dxf.handle
            self._image_def = image_def
        else:
            self.dxf.discard("image_def_handle")
            self._image_def = None

    def destroy(self) -> None:
        if not self.is_alive:
            return

        reactor = self._image_def_reactor
        if reactor and reactor.is_alive:
            image_def = self.image_def
            if image_def and image_def.is_alive:
                image_def.discard_reactor_handle(reactor.dxf.handle)
            reactor.destroy()
        super().destroy()

    def audit(self, auditor: "Auditor") -> None:
        super().audit(auditor)


# DXF reference error: Subclass marker (AcDbRasterImage)
acdb_wipeout = DefSubclass("AcDbWipeout", dict(acdb_image.attribs))
acdb_wipeout_group_codes = group_code_mapping(acdb_wipeout)


@register_entity
class Wipeout(ImageBase):
    """DXF WIPEOUT entity"""

    DXFTYPE = "WIPEOUT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_wipeout)
    DEFAULT_ATTRIBS = {
        "layer": "0",
        "flags": 7,
        "clipping": 1,
        "brightness": 50,
        "contrast": 50,
        "fade": 0,
        "image_size": (1, 1),
        "image_def_handle": "0",  # has no ImageDef()
        "image_def_reactor_handle": "0",  # has no ImageDefReactor()
        "clip_mode": 0,
    }
    _CLS_GROUP_CODES = acdb_wipeout_group_codes
    _SUBCLASS_NAME = acdb_wipeout.name  # type: ignore

    def set_masking_area(self, vertices: Iterable["Vertex"]) -> None:
        """Set a new masking area, the area is placed in the layout xy-plane."""
        self.update_dxf_attribs(self.DEFAULT_ATTRIBS)
        vertices = Vec2.list(vertices)
        bounds = BoundingBox2d(vertices)
        x_size, y_size = bounds.size

        dxf = self.dxf
        dxf.insert = Vec3(bounds.extmin)
        dxf.u_pixel = Vec3(x_size, 0, 0)
        dxf.v_pixel = Vec3(0, y_size, 0)

        def boundary_path():
            extmin = bounds.extmin
            for vertex in vertices:
                v = vertex - extmin
                yield Vec2(v.x / x_size - 0.5, 0.5 - v.y / y_size)

        self.set_boundary_path(boundary_path())

    def _reset_handles(self):
        self.dxf.image_def_reactor_handle = "0"
        self.dxf.image_def_handle = "0"

    def audit(self, auditor: "Auditor") -> None:
        self._reset_handles()
        super().audit(auditor)

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        self._reset_handles()
        super().export_entity(tagwriter)


acdb_image_def = DefSubclass(
    "AcDbRasterImageDef",
    {
        "class_version": DXFAttr(90, default=0),
        # File name of image:
        "filename": DXFAttr(1),
        # Image size in pixels:
        "image_size": DXFAttr(10, xtype=XType.point2d),
        # Default size of one pixel in AutoCAD units:
        "pixel_size": DXFAttr(11, xtype=XType.point2d, default=(0.01, 0.01)),
        "loaded": DXFAttr(280, default=1),
        # Resolution units - this enums differ from the usual drawing units,
        # units.py, same as for RasterVariables.dxf.units, but only these 3 values
        # are valid - confirmed by ODA Specs 20.4.81 IMAGEDEF:
        # 0 = No units
        # 2 = Centimeters
        # 5 = Inch
        "resolution_units": DXFAttr(
            281,
            default=0,
            validator=validator.is_one_of({0, 2, 5}),
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_image_def_group_codes = group_code_mapping(acdb_image_def)


@register_entity
class ImageDef(DXFObject):
    """DXF IMAGEDEF entity"""

    DXFTYPE = "IMAGEDEF"
    DXFATTRIBS = DXFAttributes(base_class, acdb_image_def)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_image_def_group_codes, 1)
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_image_def.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "class_version",
                "filename",
                "image_size",
                "pixel_size",
                "loaded",
                "resolution_units",
            ],
        )


acdb_image_def_reactor = DefSubclass(
    "AcDbRasterImageDefReactor",
    {
        "class_version": DXFAttr(90, default=2),
        # Handle to image:
        "image_handle": DXFAttr(330),
    },
)
acdb_image_def_reactor_group_codes = group_code_mapping(acdb_image_def_reactor)


@register_entity
class ImageDefReactor(DXFObject):
    """DXF IMAGEDEF_REACTOR entity"""

    DXFTYPE = "IMAGEDEF_REACTOR"
    DXFATTRIBS = DXFAttributes(base_class, acdb_image_def_reactor)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_image_def_reactor_group_codes, 1
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_image_def_reactor.name)
        tagwriter.write_tag2(90, self.dxf.class_version)
        tagwriter.write_tag2(330, self.dxf.image_handle)


acdb_raster_variables = DefSubclass(
    "AcDbRasterVariables",
    {
        "class_version": DXFAttr(90, default=0),
        # Frame:
        # 0 = no frame
        # 1 = show frame
        "frame": DXFAttr(
            70,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Quality:
        # 0 = draft
        # 1 = high
        "quality": DXFAttr(
            71,
            default=1,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Units:
        # 0 = None
        # 1 = mm
        # 2 = cm
        # 3 = m
        # 4 = km
        # 5 = in
        # 6 = ft
        # 7 = yd
        # 8 = mi
        "units": DXFAttr(
            72,
            default=3,
            validator=validator.is_in_integer_range(0, 9),
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_raster_variables_group_codes = group_code_mapping(acdb_raster_variables)


@register_entity
class RasterVariables(DXFObject):
    """DXF RASTERVARIABLES entity"""

    DXFTYPE = "RASTERVARIABLES"
    DXFATTRIBS = DXFAttributes(base_class, acdb_raster_variables)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_raster_variables_group_codes, 1
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_raster_variables.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "class_version",
                "frame",
                "quality",
                "units",
            ],
        )


acdb_wipeout_variables = DefSubclass(
    "AcDbWipeoutVariables",
    {
        # Display-image-frame flag:
        # 0 = No frame
        # 1 = Display frame
        "frame": DXFAttr(
            70,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_wipeout_variables_group_codes = group_code_mapping(acdb_wipeout_variables)


@register_entity
class WipeoutVariables(DXFObject):
    """DXF WIPEOUTVARIABLES entity"""

    DXFTYPE = "WIPEOUTVARIABLES"
    DXFATTRIBS = DXFAttributes(base_class, acdb_wipeout_variables)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_wipeout_variables_group_codes, 1
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_wipeout_variables.name)
        self.dxf.export_dxf_attribs(tagwriter, "frame")



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import math
from typing import TYPE_CHECKING, Tuple, Optional
import warnings

from ezdxf.lldxf import validator
from ezdxf.lldxf import const
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.enums import (
    TextEntityAlignment,
    MAP_STRING_ALIGN_TO_ENUM_ALIGN,
    MAP_TEXT_ENUM_TO_ALIGN_FLAGS,
    MAP_TEXT_ALIGN_FLAGS_TO_ENUM,
    MAP_FLAGS_TO_STRING_ALIGN,
)
from ezdxf.math import Vec3, Matrix44, NULLVEC, Z_AXIS
from ezdxf.math.transformtools import OCSTransform
from ezdxf.audit import Auditor
from ezdxf.tools.text import plain_text

from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity, elevation_to_z_axis
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, Vertex, DXFNamespace, Drawing

__all__ = ["Text", "acdb_text", "acdb_text_group_codes"]

acdb_text = DefSubclass(
    "AcDbText",
    {
        # First alignment point (in OCS):
        "insert": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # Text height
        "height": DXFAttr(
            40,
            default=2.5,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Text content as sting:
        "text": DXFAttr(
            1,
            default="",
            validator=validator.is_valid_one_line_text,
            fixer=validator.fix_one_line_text,
        ),
        # Text rotation in degrees (optional)
        "rotation": DXFAttr(50, default=0, optional=True),
        # Oblique angle in degrees, vertical = 0 deg (optional)
        "oblique": DXFAttr(51, default=0, optional=True),
        # Text style name (optional), given text style must have an entry in the
        # text-styles tables.
        "style": DXFAttr(7, default="Standard", optional=True),
        # Relative X scale factor—width (optional)
        # This value is also adjusted when fit-type text is used
        "width": DXFAttr(
            41,
            default=1,
            optional=True,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Text generation flags (optional)
        # 2 = backward (mirror-x),
        # 4 = upside down (mirror-y)
        "text_generation_flag": DXFAttr(
            71,
            default=0,
            optional=True,
            validator=validator.is_one_of({0, 2, 4, 6}),
            fixer=RETURN_DEFAULT,
        ),
        # Horizontal text justification type (optional) horizontal justification
        # 0 = Left
        # 1 = Center
        # 2 = Right
        # 3 = Aligned (if vertical alignment = 0)
        # 4 = Middle (if vertical alignment = 0)
        # 5 = Fit (if vertical alignment = 0)
        # This value is meaningful only if the value of a 72 or 73 group is nonzero
        # (if the justification is anything other than baseline/left)
        "halign": DXFAttr(
            72,
            default=0,
            optional=True,
            validator=validator.is_in_integer_range(0, 6),
            fixer=RETURN_DEFAULT,
        ),
        # Second alignment point (in OCS) (optional)
        "align_point": DXFAttr(11, xtype=XType.point3d, optional=True),
        # Elevation is a legacy feature from R11 and prior, do not use this
        # attribute, store the entity elevation in the z-axis of the vertices.
        # ezdxf does not export the elevation attribute!
        "elevation": DXFAttr(38, default=0, optional=True),
        # Thickness in extrusion direction, only supported for SHX font in
        # AutoCAD/BricsCAD (optional), can be negative
        "thickness": DXFAttr(39, default=0, optional=True),
        # Extrusion direction (optional)
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_text_group_codes = group_code_mapping(acdb_text)
acdb_text2 = DefSubclass(
    "AcDbText",
    {
        # Vertical text justification type (optional)
        # 0 = Baseline
        # 1 = Bottom
        # 2 = Middle
        # 3 = Top
        "valign": DXFAttr(
            73,
            default=0,
            optional=True,
            validator=validator.is_in_integer_range(0, 4),
            fixer=RETURN_DEFAULT,
        )
    },
)
acdb_text2_group_codes = group_code_mapping(acdb_text2)


# Formatting codes:
# %%d: '°'
# %%u in TEXT start underline formatting until next %%u or until end of line


@register_entity
class Text(DXFGraphic):
    """DXF TEXT entity"""

    DXFTYPE = "TEXT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_text, acdb_text2)
    # horizontal align values
    LEFT = 0
    CENTER = 1
    RIGHT = 2
    # vertical align values
    BASELINE = 0
    BOTTOM = 1
    MIDDLE = 2
    TOP = 3
    # text generation flags
    MIRROR_X = 2
    MIRROR_Y = 4
    BACKWARD = MIRROR_X
    UPSIDE_DOWN = MIRROR_Y

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        """Loading interface. (internal API)"""
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_text_group_codes, 2, recover=True
            )
            processor.fast_load_dxfattribs(
                dxf, acdb_text2_group_codes, 3, recover=True
            )
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ("insert", "align_point"))
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags. (internal API)"""
        super().export_entity(tagwriter)
        self.export_acdb_text(tagwriter)
        self.export_acdb_text2(tagwriter)

    def export_acdb_text(self, tagwriter: "TagWriter") -> None:
        """Export TEXT data as DXF tags. (internal API)"""
        if tagwriter.dxfversion > const.DXF12:
            tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_text.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "insert",
                "height",
                "text",
                "thickness",
                "rotation",
                "oblique",
                "style",
                "width",
                "text_generation_flag",
                "halign",
                "align_point",
                "extrusion",
            ],
        )

    def export_acdb_text2(self, tagwriter: "TagWriter") -> None:
        """Export TEXT data as DXF tags. (internal API)"""
        if tagwriter.dxfversion > const.DXF12:
            tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_text2.name)
        self.dxf.export_dxf_attribs(tagwriter, "valign")

    def set_pos(
        self, p1: "Vertex", p2: "Vertex" = None, align: str = None
    ) -> "Text":
        """Set text alignment and location. (deprecated)

        The alignments "ALIGNED" and "FIT"
        are special, they require a second alignment point, the text is aligned
        on the virtual line between these two points and sits vertically at the
        base line.

        - "ALIGNED": Text is stretched or compressed
          to fit exactly between `p1` and `p2` and the text height is also
          adjusted to preserve height/width ratio.
        - "FIT": Text is stretched or compressed to fit
          exactly between `p1` and `p2` but only the text width is adjusted,
          the text height is fixed by the :attr:`dxf.height` attribute.
        - "MIDDLE": also a special adjustment, centered
          text like "MIDDLE_CENTER", but vertically
          centred at the total height of the text.

        .. warning::

            Will be removed in v1.0.0, use :meth:`set_placement`

        Args:
            p1: first alignment point as (x, y[, z])
            p2: second alignment point as (x, y[, z]), required for "ALIGNED"
                and "FIT" else ignored
            align: new alignment as string or ``None`` to preserve the existing
                alignment.

        """
        warnings.warn(
            "set_pos() is deprecated, replaced by set_placement()",
            DeprecationWarning,
        )

        if align is None:
            align_enum = self.get_align_enum()
        else:
            assert isinstance(align, str)
            try:
                align_enum = MAP_STRING_ALIGN_TO_ENUM_ALIGN[align.upper()]
            except KeyError:
                raise ValueError(align)
            self.set_align_enum(align_enum)
        return self.set_placement(p1, p2, align_enum)

    def set_placement(
        self,
        p1: "Vertex",
        p2: "Vertex" = None,
        align: TextEntityAlignment = None,
    ) -> "Text":
        """Set text alignment and location.

        The alignments :attr:`ALIGNED` and :attr:`FIT`
        are special, they require a second alignment point, the text is aligned
        on the virtual line between these two points and sits vertically at the
        base line.

        - :attr:`ALIGNED`: Text is stretched or compressed
          to fit exactly between `p1` and `p2` and the text height is also
          adjusted to preserve height/width ratio.
        - :attr:`FIT`: Text is stretched or compressed to fit
          exactly between `p1` and `p2` but only the text width is adjusted,
          the text height is fixed by the :attr:`dxf.height` attribute.
        - :attr:`MIDDLE`: also a special adjustment, centered
          text like :attr:`MIDDLE_CENTER`, but vertically
          centred at the total height of the text.

        Args:
            p1: first alignment point as (x, y[, z])
            p2: second alignment point as (x, y[, z]), required for :attr:`ALIGNED`
                and :attr:`FIT` else ignored
            align: new alignment as enum :class:`~ezdxf.enums.TextEntityAlignment`,
                ``None`` to preserve the existing alignment.

        """
        if align is None:
            align = self.get_align_enum()
        else:
            assert isinstance(align, TextEntityAlignment)
            self.set_align_enum(align)
        self.dxf.insert = p1
        if align in (TextEntityAlignment.ALIGNED, TextEntityAlignment.FIT):
            if p2 is None:
                raise const.DXFValueError(
                    f"Alignment '{str(align)}' requires a second alignment point."
                )
        else:
            p2 = p1
        self.dxf.align_point = p2
        return self

    def get_pos(self) -> Tuple[str, Vec3, Optional[Vec3]]:
        """Returns a tuple (`align`, `p1`, `p2`), `align` is the alignment
        method, `p1` is the alignment point, `p2` is only relevant if `align`
        is "ALIGNED" or "FIT", otherwise it is ``None`` (deprecated).

        .. warning::

            Will be removed in v1.0.0, use :meth:`get_placement()`

        """
        warnings.warn(
            "get_pos() is deprecated, replaced by get_placement()",
            DeprecationWarning,
        )

        p1 = Vec3(self.dxf.insert)
        # Except for "LEFT" is the "align point" the real insert point:
        # If the required "align point" is not present use "insert"!
        p2 = Vec3(self.dxf.get("align_point", p1))
        align = self.get_align()
        if align == "LEFT":
            return align, p1, None
        if align in ("FIT", "ALIGNED"):
            return align, p1, p2
        return align, p2, None

    def get_placement(self) -> Tuple[TextEntityAlignment, Vec3, Optional[Vec3]]:
        """Returns a tuple (`align`, `p1`, `p2`), `align` is the alignment
        enum :class:`~ezdxf.enum.TextEntityAlignment`, `p1` is the
        alignment point, `p2` is only relevant if `align` is :attr:`ALIGNED` or
        :attr:`FIT`, otherwise it is ``None``.

        """
        p1 = Vec3(self.dxf.insert)
        # Except for "LEFT" is the "align point" the real insert point:
        # If the required "align point" is not present use "insert"!
        p2 = Vec3(self.dxf.get("align_point", p1))
        align = self.get_align_enum()
        if align is TextEntityAlignment.LEFT:
            return align, p1, None
        if align in (TextEntityAlignment.FIT, TextEntityAlignment.ALIGNED):
            return align, p1, p2
        return align, p2, None

    def set_align(self, align: str = "LEFT") -> "Text":
        """Set the text alignment as string (deprecated)

        .. warning::

            Will be removed in v1.0.0, use :meth:`set_align_enum`

        """
        warnings.warn(
            "set_align() is deprecated, replaced by set_align_enum()",
            DeprecationWarning,
        )
        assert isinstance(align, str)
        align = align.upper()
        try:
            align_enum = MAP_STRING_ALIGN_TO_ENUM_ALIGN[align]
        except KeyError:
            raise ValueError(f"invalid argument align: {align}")
        halign, valign = MAP_TEXT_ENUM_TO_ALIGN_FLAGS[align_enum]
        self.dxf.halign = halign
        self.dxf.valign = valign
        return self

    def set_align_enum(self, align=TextEntityAlignment.LEFT) -> "Text":
        """Just for experts: Sets the text alignment without setting the
        alignment points, set adjustment points attr:`dxf.insert` and
        :attr:`dxf.align_point` manually.

        Args:
            align: :class:`~ezdxf.enums.TextEntityAlignment`

        """
        halign, valign = MAP_TEXT_ENUM_TO_ALIGN_FLAGS[align]
        self.dxf.halign = halign
        self.dxf.valign = valign
        return self

    def get_align(self) -> str:
        """Returns the current text alignment as string (deprecated).

        .. warning::

            Will be removed in v1.0.0, use :meth:`get_align_enum`

        """
        warnings.warn(
            "get_align() is deprecated, replaced by get_align_enum()",
            DeprecationWarning,
        )

        halign = self.dxf.get("halign", 0)
        valign = self.dxf.get("valign", 0)
        if halign > 2:
            valign = 0
        return MAP_FLAGS_TO_STRING_ALIGN.get((halign, valign), "LEFT")

    def get_align_enum(self) -> TextEntityAlignment:
        """Returns the current text alignment as :class:`~ezdxf.enums.TextEntityAlignment`,
        see also :meth:`set_placement`.
        """
        halign = self.dxf.get("halign", 0)
        valign = self.dxf.get("valign", 0)
        if halign > 2:
            valign = 0
        return MAP_TEXT_ALIGN_FLAGS_TO_ENUM.get(
            (halign, valign), TextEntityAlignment.LEFT
        )

    def transform(self, m: Matrix44) -> "Text":
        """Transform the TEXT entity by transformation matrix `m` inplace."""
        dxf = self.dxf
        if not dxf.hasattr("align_point"):
            dxf.align_point = dxf.insert
        ocs = OCSTransform(self.dxf.extrusion, m)
        dxf.insert = ocs.transform_vertex(dxf.insert)
        dxf.align_point = ocs.transform_vertex(dxf.align_point)
        old_rotation = dxf.rotation
        new_rotation = ocs.transform_deg_angle(old_rotation)
        x_scale = ocs.transform_length(Vec3.from_deg_angle(old_rotation))
        y_scale = ocs.transform_length(Vec3.from_deg_angle(old_rotation + 90.0))

        if not ocs.scale_uniform:
            oblique_vec = Vec3.from_deg_angle(old_rotation + 90.0 - dxf.oblique)
            new_oblique_deg = (
                new_rotation
                + 90.0
                - ocs.transform_direction(oblique_vec).angle_deg
            )
            dxf.oblique = new_oblique_deg
            y_scale *= math.cos(math.radians(new_oblique_deg))

        dxf.width *= x_scale / y_scale
        dxf.height *= y_scale
        dxf.rotation = new_rotation

        if dxf.hasattr("thickness"):  # can be negative
            dxf.thickness = ocs.transform_thickness(dxf.thickness)
        dxf.extrusion = ocs.new_extrusion
        self.post_transform(m)
        return self

    def translate(self, dx: float, dy: float, dz: float) -> "Text":
        """Optimized TEXT/ATTRIB/ATTDEF translation about `dx` in x-axis, `dy`
        in y-axis and `dz` in z-axis, returns `self`.

        """
        ocs = self.ocs()
        dxf = self.dxf
        vec = Vec3(dx, dy, dz)

        dxf.insert = ocs.from_wcs(vec + ocs.to_wcs(dxf.insert))
        if dxf.hasattr("align_point"):
            dxf.align_point = ocs.from_wcs(vec + ocs.to_wcs(dxf.align_point))
        # Avoid Matrix44 instantiation if not required:
        if self.is_post_transform_required:
            self.post_transform(Matrix44.translate(dx, dy, dz))
        return self

    def remove_dependencies(self, other: "Drawing" = None) -> None:
        """Remove all dependencies from actual document.

        (internal API)
        """
        if not self.is_alive:
            return

        super().remove_dependencies()
        has_style = other is not None and (self.dxf.style in other.styles)
        if not has_style:
            self.dxf.style = "Standard"

    def plain_text(self) -> str:
        """Returns text content without formatting codes."""
        return plain_text(self.dxf.text)

    def audit(self, auditor: Auditor):
        """Validity check."""
        super().audit(auditor)
        auditor.check_text_style(self)

    @property
    def is_backward(self) -> bool:
        """Get/set text generation flag BACKWARDS, for mirrored text along the
        x-axis.
        """
        return bool(self.dxf.text_generation_flag & const.BACKWARD)

    @is_backward.setter
    def is_backward(self, state) -> None:
        self.set_flag_state(const.BACKWARD, state, "text_generation_flag")

    @property
    def is_upside_down(self) -> bool:
        """Get/set text generation flag UPSIDE_DOWN, for mirrored text along
        the y-axis.

        """
        return bool(self.dxf.text_generation_flag & const.UPSIDE_DOWN)

    @is_upside_down.setter
    def is_upside_down(self, state) -> None:
        self.set_flag_state(const.UPSIDE_DOWN, state, "text_generation_flag")

    def wcs_transformation_matrix(self) -> Matrix44:
        return text_transformation_matrix(self)

    def font_name(self) -> str:
        """Returns the font name of the associated :class:`Textstyle`."""
        font_name = "arial.ttf"
        style_name = self.dxf.style
        if self.doc:
            try:
                style = self.doc.styles.get(style_name)
                font_name = style.dxf.font
            except ValueError:
                pass
        return font_name

    def fit_length(self) -> float:
        """Returns the text length for alignments :attr:`TextEntityAlignment.FIT`
        and :attr:`TextEntityAlignment.ALIGNED`, defined by the distance from
        the insertion point to the align point or 0 for all other alignments.

        """
        length = 0.0
        align, p1, p2 = self.get_placement()
        if align in (TextEntityAlignment.FIT, TextEntityAlignment.ALIGNED):
            # text is stretch between p1 and p2
            length = p1.distance(p2)
        return length


def text_transformation_matrix(entity: Text) -> Matrix44:
    """Apply rotation, width factor, translation to the insertion point
    and if necessary transformation from OCS to WCS.
    """
    angle = math.radians(entity.dxf.rotation)
    width_factor = entity.dxf.width
    align, p1, p2 = entity.get_placement()
    mirror_x = -1 if entity.is_backward else 1
    mirror_y = -1 if entity.is_upside_down else 1
    oblique = math.radians(entity.dxf.oblique)
    location = p1
    if align in (TextEntityAlignment.ALIGNED, TextEntityAlignment.FIT):
        width_factor = 1.0  # text goes from p1 to p2, no stretching applied
        location = p1.lerp(p2, factor=0.5)
        angle = (p2 - p1).angle  # override stored angle

    m = Matrix44()
    if oblique:
        m *= Matrix44.shear_xy(angle_x=oblique)
    sx = width_factor * mirror_x
    sy = mirror_y
    if sx != 1 or sy != 1:
        m *= Matrix44.scale(sx, sy, 1)
    if angle:
        m *= Matrix44.z_rotate(angle)
    if location:
        m *= Matrix44.translate(location.x, location.y, location.z)

    ocs = entity.ocs()
    if ocs.transform:  # to WCS
        m *= ocs.matrix
    return m



# Copyright (c) 2018-2021, Manfred Moitzi
# License: MIT License
from typing import (
    TYPE_CHECKING,
    Tuple,
    Iterable,
    Optional,
    Dict,
    Any,
    List,
    cast,
)
import math
import abc
from ezdxf.math import (
    Vec3,
    Vec2,
    ConstructionLine,
    ConstructionBox,
    ConstructionArc,
)
from ezdxf.math import UCS, PassTroughUCS, xround, Z_AXIS
from ezdxf.lldxf import const
from ezdxf.enums import TextEntityAlignment
from ezdxf._options import options
from ezdxf.lldxf.const import DXFValueError, DXFUndefinedBlockError
from ezdxf.tools import suppress_zeros
from ezdxf.render.arrows import ARROWS
from ezdxf.entities import DimStyleOverride, Dimension

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        Vertex,
        Drawing,
        GenericLayoutType,
        Textstyle,
    )


class TextBox(ConstructionBox):
    """Text boundaries representation."""

    def __init__(
        self,
        center: Vec2 = Vec2(0, 0),
        width: float = 0.0,
        height: float = 0.0,
        angle: float = 0.0,
        hgap: float = 0.0,  # horizontal gap - width
        vgap: float = 0.0,  # vertical gap - height
    ):
        super().__init__(
            center, width + 2.0 * hgap, height + 2.0 * vgap, angle
        )


PLUS_MINUS = "±"
_TOLERANCE_COMMON = r"\A{align};{txt}{{\H{fac:.2f}x;"
TOLERANCE_TEMPLATE1 = _TOLERANCE_COMMON + r"{tol}}}"
TOLERANCE_TEMPLATE2 = _TOLERANCE_COMMON + r"\S{upr}^ {lwr};}}"
LIMITS_TEMPLATE = r"{{\H{fac:.2f}x;\S{upr}^ {lwr};}}"


def OptionalVec2(v) -> Optional[Vec2]:
    if v is not None:
        return Vec2(v)
    else:
        return None


def sign_char(value: float) -> str:
    if value < 0.0:
        return "-"
    elif value > 0:
        return "+"
    else:
        return " "


def format_text(
    value: float,
    dimrnd: float = None,
    dimdec: int = 2,
    dimzin: int = 0,
    dimdsep: str = ".",
) -> str:
    if dimrnd is not None:
        value = xround(value, dimrnd)

    if dimdec is None:
        fmt = "{:f}"
        # Remove pending zeros for undefined decimal places:
        # '{:f}'.format(0) -> '0.000000'
        dimzin = dimzin | 8
    else:
        fmt = "{:." + str(dimdec) + "f}"
    text = fmt.format(value)

    leading = bool(dimzin & 4)
    pending = bool(dimzin & 8)
    text = suppress_zeros(text, leading, pending)
    if dimdsep != ".":
        text = text.replace(".", dimdsep)
    return text


def apply_dimpost(text: str, dimpost: str) -> str:
    if "<>" in dimpost:
        fmt = dimpost.replace("<>", "{}", 1)
        return fmt.format(text)
    else:
        raise DXFValueError(f'Invalid dimpost string: "{dimpost}"')


class Tolerance:  # and Limits
    def __init__(
        self,
        dim_style: DimStyleOverride,
        cap_height: float = 1.0,
        width_factor: float = 1.0,
        dim_scale: float = 1.0,
    ):
        self.text_width_factor = width_factor
        self.dim_scale = dim_scale
        get = dim_style.get
        # Appends tolerances to dimension text.
        # enabling DIMTOL disables DIMLIM.
        self.has_tolerance = bool(get("dimtol", 0))
        self.has_limits = False
        if not self.has_tolerance:
            # Limits generates dimension limits as the default text.
            self.has_limits = bool(get("dimlim", 0))

        # Scale factor for the text height of fractions and tolerance values
        # relative to the dimension text height
        self.text_scale_factor: float = get("dimtfac", 0.5)

        self.text_decimal_separator = dim_style.get_decimal_separator()

        # Default MTEXT line spacing for tolerances (BricsCAD)
        self.line_spacing: float = 1.35

        # Sets the minimum (or lower) tolerance limit for dimension text when
        # DIMTOL or DIMLIM is on.
        # DIMTM accepts signed values.
        # If DIMTOL is on and DIMTP and DIMTM are set to the same value, a
        # tolerance value is drawn.
        # If DIMTM and DIMTP values differ, the upper tolerance is drawn above
        # the lower, and a plus sign is added to the DIMTP value if it is
        # positive.
        # For DIMTM, the program uses the negative of the value you enter
        # (adding a minus sign if you specify a positive number and a plus sign
        # if you specify a negative number).
        self.minimum: float = get("dimtm", 0.0)

        # Sets the maximum (or upper) tolerance limit for dimension text when
        # DIMTOL or DIMLIM is on.
        # DIMTP accepts signed values.
        # If DIMTOL is on and DIMTP and DIMTM are set to the same value, a
        # tolerance value is drawn.
        # If DIMTM and DIMTP values differ, the upper tolerance is drawn above
        # the lower and a plus sign is added to the DIMTP value if it is
        # positive.
        self.maximum: float = get("dimtp", 0.0)

        # Number of decimal places to display in tolerance values
        # Same value for linear and angular measurements!
        self.decimal_places: int = get("dimtdec", 2)

        # Vertical justification for tolerance values relative to the nominal dimension text
        # 0 = Bottom
        # 1 = Middle
        # 2 = Top
        self.valign: int = get("dimtolj", 0)

        # Same as DIMZIN for tolerances (self.text_suppress_zeros)
        # Same value for linear and angular measurements!
        self.suppress_zeros: int = get("dimtzin", 0)
        self.text: str = ""
        self.text_height: float = 0.0
        self.text_width: float = 0.0
        self.text_upper: str = ""
        self.text_lower: str = ""
        self.char_height: float = (
            cap_height * self.text_scale_factor * self.dim_scale
        )
        if self.has_tolerance:
            self.init_tolerance()
        elif self.has_limits:
            self.init_limits()

    @property
    def enabled(self) -> bool:
        return self.has_tolerance or self.has_limits

    def disable(self):
        self.has_tolerance = False
        self.has_limits = False

    def init_tolerance(self):
        # The tolerance values are stored in the dimension style, they are
        # independent from the actual measurement:
        # Single tolerance value +/- value
        if self.minimum == self.maximum:
            self.text_height = self.char_height
            self.text_width = self.get_text_width(self.text, self.text)
        else:  # 2 stacked values: +upper tolerance <above> -lower tolerance
            # requires 2 text lines
            self.text_height = self.char_height + (
                self.text_height * self.line_spacing
            )
            self.text_width = self.get_text_width(
                self.text_upper, self.text_lower
            )
        self.update_tolerance_text(self.maximum, self.minimum)

    def update_tolerance_text(self, tol_upper: float, tol_lower: float):
        if tol_upper == tol_lower:
            self.text = PLUS_MINUS + self.format_text(abs(tol_upper))
        else:
            self.text_upper = sign_char(tol_upper) + self.format_text(
                abs(tol_upper)
            )
            self.text_lower = sign_char(tol_lower * -1) + self.format_text(
                abs(tol_lower)
            )

    def init_limits(self):
        # self.text is always an empty string (default value)
        # Limit text are always 2 stacked numbers and requires the actual
        # measurement!
        self.text_height = self.char_height + (
            self.text_height * self.line_spacing
        )

    def format_text(self, value: float) -> str:
        """Rounding and text formatting of tolerance `value`, removes leading
        and trailing zeros if necessary.

        """
        # dimpost is not applied to limits or tolerances!
        return format_text(
            value=value,
            dimrnd=None,
            dimdec=self.decimal_places,
            dimzin=self.suppress_zeros,
            dimdsep=self.text_decimal_separator,
        )

    def get_text_width(self, upr: str, lwr: str) -> float:
        """Returns the text width of the tolerance (upr/lwr) in drawing units."""
        # todo: use matplotlib support
        count = max(len(upr), len(lwr))
        return self.text_height * self.text_width_factor * count

    def compile_mtext(self, text: str) -> str:
        if self.has_tolerance:
            align = max(int(self.valign), 0)
            align = min(align, 2)
            if not self.text:
                text = TOLERANCE_TEMPLATE2.format(
                    align=align,
                    txt=text,
                    fac=self.text_scale_factor,
                    upr=self.text_upper,
                    lwr=self.text_lower,
                )
            else:
                text = TOLERANCE_TEMPLATE1.format(
                    align=align,
                    txt=text,
                    fac=self.text_scale_factor,
                    tol=self.text,
                )
        elif self.has_limits:
            text = LIMITS_TEMPLATE.format(
                upr=self.text_upper,
                lwr=self.text_lower,
                fac=self.text_scale_factor,
            )
        return text

    def update_limits(self, measurement: float) -> None:
        upper_limit = measurement + self.maximum
        lower_limit = measurement - self.minimum
        self.text_upper = self.format_text(upper_limit)
        self.text_lower = self.format_text(lower_limit)
        self.text_width = self.get_text_width(self.text_upper, self.text_lower)


class ExtensionLines:
    default_lineweight: int = const.LINEWEIGHT_BYBLOCK

    def __init__(
        self, dim_style: DimStyleOverride, default_color: int, scale: float
    ):
        get = dim_style.get
        self.color: int = get("dimclre", default_color)  # ACI
        self.linetype1: str = get("dimltex1", "")
        self.linetype2: str = get("dimltex2", "")
        self.lineweight: int = get("dimlwe", self.default_lineweight)
        self.suppress1: bool = bool(get("dimse1", 0))
        self.suppress2: bool = bool(get("dimse2", 0))

        # Extension of extension line above the dimension line, in extension
        # line direction in most cases perpendicular to dimension line
        # (oblique!)
        self.extension_above: float = get("dimexe", 0.0) * scale

        # Distance of extension line from the measurement point in extension
        # line direction
        self.offset: float = get("dimexo", 0.0) * scale

        # Fixed length extension line, lenght above dimension line is still
        # self.ext_line_extension
        self.has_fixed_length: bool = bool(get("dimfxlon", 0))

        # Length below the dimension line:
        self.length_below: float = get("dimfxl", self.extension_above) * scale

    def dxfattribs(self, num: int = 1) -> Dict[str, Any]:
        """Returns default dimension line DXF attributes as dict."""
        attribs: Dict[str, Any] = {"color": self.color}
        if num == 1:
            linetype = self.linetype1
        elif num == 2:
            linetype = self.linetype2
        else:
            raise ValueError(f"invalid argument num:{num}")

        if linetype:
            attribs["linetype"] = linetype
        if self.lineweight != self.default_lineweight:
            attribs["lineweight"] = self.lineweight
        return attribs


class DimensionLine:
    default_lineweight: int = const.LINEWEIGHT_BYBLOCK

    def __init__(
        self, dim_style: DimStyleOverride, default_color: int, scale: float
    ):
        get = dim_style.get
        self.color: int = get("dimclrd", default_color)  # ACI

        # Dimension line extension, along the dimension line direction ('left'
        # and 'right')
        self.extension: float = get("dimdle", 0.0) * scale
        self.linetype: str = get("dimltype", "")
        self.lineweight: int = get("dimlwd", self.default_lineweight)

        # Suppress first part of the dimension line
        self.suppress1: bool = bool(get("dimsd1", 0))

        # Suppress second part of the dimension line
        self.suppress2: bool = bool(get("dimsd2", 0))

        # Controls whether a dimension line is drawn between the extension lines
        # even when the text is placed outside.
        # For radius and diameter dimensions (when DIMTIX is off), draws a
        # dimension line inside the circle or arc and places the text,
        # arrowheads, and leader outside.
        # 0 = no dimension line
        # 1 = draw dimension line
        # not supported yet - ezdxf behaves like option 1
        self.has_dim_line_if_text_outside: bool = bool(get("dimtofl", 1))

    def dxfattribs(self) -> Dict[str, Any]:
        """Returns default dimension line DXF attributes as dict."""
        attribs: Dict[str, Any] = {"color": self.color}
        if self.linetype:
            attribs["linetype"] = self.linetype
        if self.lineweight != self.default_lineweight:
            attribs["lineweight"] = self.lineweight
        return attribs


class Arrows:
    def __init__(self, dim_style: DimStyleOverride, color: int, scale: float):
        get = dim_style.get
        self.color: int = color
        self.tick_size: float = get("dimtsz", 0.0) * scale
        self.arrow1_name: str = ""  # empty string is a closed filled arrow
        self.arrow2_name: str = ""  # empty string is a closed filled arrow
        self.arrow_size: float = get("dimasz", 0.25) * scale
        self.suppress1 = False  # ezdxf only
        self.suppress2 = False  # ezdxf only

        if self.tick_size > 0.0:
            # Use oblique strokes as 'arrows', disables usual 'arrows' and user
            # defined blocks tick size is per definition double the size of
            # arrow size adjust arrow size to reuse the 'oblique' arrow block
            self.arrow_size = self.tick_size * 2.0
        else:
            # Arrow name or block name if user defined arrow
            (
                self.arrow1_name,
                self.arrow2_name,
            ) = dim_style.get_arrow_names()

    @property
    def has_ticks(self) -> bool:
        return self.tick_size > 0.0

    def dxfattribs(self) -> Dict[str, Any]:
        return {"color": self.color}


class Measurement:
    def __init__(
        self,
        dim_style: DimStyleOverride,
        color: int,
        scale: float,
    ):
        # update this values in method Measurement.update()
        # raw measured value
        self.raw_value: float = 0.0
        # scaled measured value
        self.value: float = 0.0
        # Final formatted dimension text
        self.text: str = ""

        dimension = dim_style.dimension
        doc = dimension.doc
        assert doc is not None, "valid DXF document required"

        # ezdxf specific attributes beyond DXF reference, therefore not stored
        # in the DXF file (DSTYLE).
        # Some of these are just an rendering effect, which will be ignored by
        # CAD applications if they modify the DIMENSION entity

        # User location override as UCS coordinates, stored as text_midpoint in
        # the DIMENSION entity
        self.user_location: Optional[Vec2] = OptionalVec2(
            dim_style.pop("user_location", None)
        )

        # User location override relative to dimline center if True
        self.relative_user_location: bool = dim_style.pop(
            "relative_user_location", False
        )

        # Shift text away from default text location - implemented as user
        # location override without leader
        # Shift text along in text direction:
        self.text_shift_h: float = dim_style.pop("text_shift_h", 0.0)
        # Shift text perpendicular to text direction:
        self.text_shift_v: float = dim_style.pop("text_shift_v", 0.0)
        # End of ezdxf specific attributes

        get = dim_style.get
        # ezdxf locates attachment points always in the text center.
        # Fixed predefined value for ezdxf rendering:
        self.text_attachment_point: int = 5

        # Ignored by ezdxf:
        self.horizontal_direction: Optional[float] = dimension.get_dxf_attrib(
            "horizontal_direction", None
        )

        # Dimension measurement factor:
        self.measurement_factor: float = get("dimlfac", 1.0)

        # Text style
        style_name: str = get("dimtxsty", options.default_dimension_text_style)
        if style_name not in doc.tables.styles:
            style_name = "Standard"
        self.text_style_name: str = style_name
        text_style = get_text_style(doc, style_name)
        self.text_height: float = get_char_height(dim_style, text_style) * scale
        self.text_width_factor: float = text_style.get_dxf_attrib("width", 1.0)
        self.stored_dim_text = dimension.dxf.text

        # text_gap: gap between dimension line an dimension text
        self.text_gap: float = get("dimgap", 0.625) * scale

        # User defined text rotation - overrides everything:
        self.user_text_rotation: float = dimension.get_dxf_attrib(
            "text_rotation", None
        )
        # calculated text rotation
        self.text_rotation: float = self.user_text_rotation
        self.text_color: int = get("dimclrt", color)  # ACI
        self.text_round: Optional[float] = get("dimrnd", None)
        self.decimal_places: int = get("dimdec", 2)
        self.angular_decimal_places: int = get("dimadec", 2)

        # Controls the suppression of zeros in the primary unit value.
        # Values 0-3 affect feet-and-inch dimensions only and are not supported
        # 4 (Bit 3) = Suppresses leading zeros in decimal dimensions,
        #   e.g. 0.5000 becomes .5000
        # 8 (Bit 4) = Suppresses trailing zeros in decimal dimensions,
        #   e.g. 12.5000 becomes 12.5
        # 12 (Bit 3+4) = Suppresses both leading and trailing zeros,
        #   e.g. 0.5000 becomes .5)
        self.suppress_zeros: int = get("dimzin", 8)

        # special setting for angular dimensions  (dimzin << 2) & 3
        # 0 = Displays all leading and trailing zeros
        # 1 = Suppresses leading zeros (for example, 0.5000 becomes .5000)
        # 2 = Suppresses trailing zeros (for example, 12.5000 becomes 12.5)
        # 3 = Suppresses leading and trailing zeros (for example, 0.5000 becomes .5)
        self.angular_suppress_zeros: int = get("dimazin", 2)

        # decimal separator char, default is ",":
        self.decimal_separator: str = dim_style.get_decimal_separator()

        self.text_post_process_format: str = get("dimpost", "")
        # text_fill:
        # 0 = None
        # 1 = Background
        # 2 = DIMTFILLCLR
        self.text_fill: int = get("dimtfill", 0)
        self.text_fill_color: int = get("dimtfillclr", 1)  # ACI
        self.text_box_fill_scale: float = 1.1

        # text_halign:
        # 0 = center
        # 1 = left
        # 2 = right
        # 3 = above ext1
        # 4 = above ext2
        self.text_halign: int = get("dimjust", 0)

        # text_valign:
        # 0 = center
        # 1 = above
        # 2 = farthest away?
        # 3 = JIS;
        # 4 = below
        # Options 2, 3 are ignored by ezdxf
        self.text_valign: int = get("dimtad", 0)

        # Controls the vertical position of dimension text above or below the
        # dimension line, when DIMTAD = 0.
        # The magnitude of the vertical offset of text is the product of the
        # text height (+gap?) and DIMTVP.
        # Setting DIMTVP to 1.0 is equivalent to setting DIMTAD = 1.
        self.text_vertical_position: float = get("dimtvp", 0.0)

        # Move text freely:
        # 0 = Moves the dimension line with dimension text
        # 1 = Adds a leader when dimension text is moved
        # 2 = Allows text to be moved freely without a leader
        self.text_movement_rule: int = get("dimtmove", 2)

        self.has_leader: bool = (
            self.user_location is not None and self.text_movement_rule == 1
        )

        # text_rotation is 0 if dimension text is 'inside', ezdxf defines
        # 'inside' as at the default text location:
        self.text_inside_horizontal: bool = get("dimtih", 0)

        # text_rotation is 0 if dimension text is 'outside', ezdxf defines
        # 'outside' as NOT at the default text location:
        self.text_outside_horizontal: bool = get("dimtoh", 0)

        # Force text location 'inside', even if the text should be moved
        # 'outside':
        self.force_text_inside: bool = bool(get("dimtix", 0))

        # How dimension text and arrows are arranged when space is not
        # sufficient to place both 'inside':
        # 0 = Places both text and arrows outside extension lines
        # 1 = Moves arrows first, then text
        # 2 = Moves text first, then arrows
        # 3 = Moves either text or arrows, whichever fits best
        # not supported - ezdxf behaves like 2
        self.text_fitting_rule: int = get("dimatfit", 2)

        # Units for all dimension types except Angular.
        # 1 = Scientific
        # 2 = Decimal
        # 3 = Engineering
        # 4 = Architectural (always displayed stacked)
        # 5 = Fractional (always displayed stacked)
        # not supported - ezdxf behaves like 2
        self.length_unit: int = get("dimlunit", 2)

        # Fraction format when DIMLUNIT is set to 4 (Architectural) or
        # 5 (Fractional).
        # 0 = Horizontal stacking
        # 1 = Diagonal stacking
        # 2 = Not stacked (for example, 1/2)
        self.fraction_format: int = get("dimfrac", 0)  # not supported

        # Units format for angular dimensions
        # 0 = Decimal degrees
        # 1 = Degrees/minutes/seconds
        # 2 = Grad
        # 3 = Radians
        self.angle_units: int = get("dimaunit", 0)

        self.has_arc_length_prefix = False
        if get("dimarcsym", 2) == 0:
            self.has_arc_length_prefix = True

        # Text_outside is only True if really placed outside of default text
        # location
        # remark: user defined text location is always outside per definition
        # (not by real location)
        self.text_is_outside: bool = False

        # Final calculated or overridden dimension text location
        self.text_location: Vec2 = Vec2(0, 0)

        # True if dimension text doesn't fit between extension lines
        self.is_wide_text: bool = False

        # Text rotation was corrected to make upside down text better readable
        self.has_upside_down_correction: bool = False

    @property
    def text_is_inside(self):
        return not self.text_is_outside

    @property
    def has_relative_text_movement(self):
        return bool(self.text_shift_h or self.text_shift_v)

    def apply_text_shift(self, location: Vec2, text_rotation: float) -> Vec2:
        """Add `self.text_shift_h` and `sel.text_shift_v` to point `location`,
        shifting along and perpendicular to text orientation defined by
        `text_rotation`.

        Args:
            location: location point
            text_rotation: text rotation in degrees

        Returns: new location

        """
        shift_vec = Vec2((self.text_shift_h, self.text_shift_v))
        location += shift_vec.rotate_deg(text_rotation)
        return location

    @property
    def vertical_placement(self) -> float:
        """Returns vertical placement of dimension text as 1 for above, 0 for
        center and -1 for below dimension line.

        """
        if self.text_valign == 0:
            return 0
        elif self.text_valign == 4:
            return -1
        else:
            return 1

    def text_vertical_distance(self) -> float:
        """Returns the vertical distance for dimension line to text midpoint.
        Positive values are above the line, negative values are below the line.

        """
        if self.text_valign == 0:
            return self.text_height * self.text_vertical_position
        else:
            return (
                self.text_height / 2.0 + self.text_gap
            ) * self.vertical_placement

    def text_width(self, text: str) -> float:
        """
        Return width of `text` in drawing units.

        """
        # todo: use matplotlib support
        char_width = self.text_height * self.text_width_factor
        return len(text) * char_width

    def text_override(self, measurement: float) -> str:
        """Create dimension text for `measurement` in drawing units and applies
        text overriding properties.

        """
        text = self.stored_dim_text
        if text == " ":  # suppresses text
            return ""
        elif text == "" or text == "<>":  # measured distance
            return self.format_text(measurement)
        else:  # user override
            return text

    def location_override(
        self, location: "Vertex", leader=False, relative=False
    ) -> None:
        """Set user defined dimension text location. ezdxf defines a user
        defined location per definition as 'outside'.

        Args:
            location: text midpoint
            leader: use leader or not (movement rules)
            relative: is location absolute (in UCS) or relative to dimension
                line center.

        """
        self.user_location = Vec2(location)
        self.text_movement_rule = 1 if leader else 2
        self.relative_user_location = relative
        self.text_is_outside = True

    def dxfattribs(self) -> Dict[str, Any]:
        return {"color": self.text_color}

    @abc.abstractmethod
    def update(self, raw_measurement_value: float) -> None:
        """Update raw measurement value, scaled measurement value and
        dimension text.

        """

    @abc.abstractmethod
    def format_text(self, value: float) -> str:
        """Rounding and text formatting of `value`, removes leading and
        trailing zeros if necessary.

        """


class LengthMeasurement(Measurement):
    def update(self, raw_measurement_value: float) -> None:
        """Update raw measurement value, scaled measurement value and
        dimension text.
        """
        self.raw_value = raw_measurement_value
        self.value = raw_measurement_value * self.measurement_factor
        self.text = self.text_override(self.value)

    def format_text(self, value: float) -> str:
        """Rounding and text formatting of `value`, removes leading and
        trailing zeros if necessary.

        """
        text = format_text(
            value,
            self.text_round,
            self.decimal_places,
            self.suppress_zeros,
            self.decimal_separator,
        )
        if self.text_post_process_format:
            text = apply_dimpost(text, self.text_post_process_format)
        return text


class Geometry:
    """
    Geometry layout entities are located in the OCS defined by the extrusion
    vector of the DIMENSION entity and the z-axis of the OCS
    point 'text_midpoint' (group code 11).

    """

    def __init__(
        self,
        dimension: Dimension,
        ucs: UCS,
        layout: "GenericLayoutType",
    ):
        assert dimension.doc is not None, "valid DXF document required"
        self.dimension: Dimension = dimension
        self.doc: "Drawing" = dimension.doc
        self.dxfversion: str = self.doc.dxfversion
        self.supports_dxf_r2000: bool = self.dxfversion >= "AC1015"
        self.supports_dxf_r2007: bool = self.dxfversion >= "AC1021"
        self.ucs: UCS = ucs
        self.extrusion: Vec3 = ucs.uz
        self.requires_extrusion: bool = not self.extrusion.isclose(Z_AXIS)
        self.layout: "GenericLayoutType" = layout
        self._text_box: TextBox = TextBox()

    @property
    def has_text_box(self) -> bool:
        return self._text_box.width > 0.0 and self._text_box.height > 0.0

    def set_layout(self, layout: "GenericLayoutType") -> None:
        self.layout = layout

    def set_text_box(self, text_box: TextBox) -> None:
        self._text_box = text_box

    def has_block(self, name: str) -> bool:
        return name in self.doc.blocks

    def add_arrow_blockref(
        self,
        name: str,
        insert: Vec2,
        size: float,
        rotation: float,
        dxfattribs: Dict[str, Any],
    ) -> None:
        # OCS of the arrow blocks is defined by the DIMENSION entity!
        # Therefore remove OCS elevation, the elevation is defined by the
        # DIMENSION 'text_midpoint' (group code 11) and do not set 'extrusion'
        # either!
        insert = self.ucs.to_ocs(Vec3(insert)).vec2
        rotation = self.ucs.to_ocs_angle_deg(rotation)
        self.layout.add_arrow_blockref(name, insert, size, rotation, dxfattribs)

    def add_blockref(
        self,
        name: str,
        insert: Vec2,
        rotation: float,
        dxfattribs: Dict[str, Any],
    ) -> None:
        # OCS of the arrow blocks is defined by the DIMENSION entity!
        # Therefore remove OCS elevation, the elevation is defined by the
        # DIMENSION 'text_midpoint' (group code 11) and do not set 'extrusion'
        # either!
        insert = self.ucs.to_ocs(Vec3(insert)).vec2
        dxfattribs["rotation"] = self.ucs.to_ocs_angle_deg(rotation)
        self.layout.add_blockref(name, insert, dxfattribs)

    def add_text(
        self, text: str, pos: Vec2, rotation: float, dxfattribs: Dict[str, Any]
    ) -> None:
        dxfattribs["rotation"] = self.ucs.to_ocs_angle_deg(rotation)
        entity = self.layout.add_text(text, dxfattribs)
        # OCS of the measurement text is defined by the DIMENSION entity!
        # Therefore remove OCS elevation, the elevation is defined by the
        # DIMENSION 'text_midpoint' (group code 11) and do not set 'extrusion'
        # either!
        entity.set_placement(self.ucs.to_ocs(Vec3(pos)).vec2, align=TextEntityAlignment.MIDDLE_CENTER)

    def add_mtext(
        self, text: str, pos: Vec2, rotation: float, dxfattribs: Dict[str, Any]
    ) -> None:
        # OCS of the measurement text is defined by the DIMENSION entity!
        # Therefore remove OCS elevation, the elevation is defined by the
        # DIMENSION 'text_midpoint' (group code 11) and do not set 'extrusion'
        # either!
        dxfattribs["rotation"] = self.ucs.to_ocs_angle_deg(rotation)
        dxfattribs["insert"] = self.ucs.to_ocs(Vec3(pos)).vec2
        self.layout.add_mtext(text, dxfattribs)

    def add_defpoints(self, points: Iterable[Vec2]) -> None:
        attribs = {
            "layer": "Defpoints",
        }
        for point in points:
            # Despite the fact that the POINT entity has WCS coordinates,
            # the coordinates of defpoints in DIMENSION entities have OCS
            # coordinates.
            location = self.ucs.to_ocs(Vec3(point)).replace(z=0)
            self.layout.add_point(location, dxfattribs=attribs)

    def add_line(
        self,
        start: Vec2,
        end: Vec2,
        dxfattribs: Dict[str, Any],
        remove_hidden_lines=False,
    ) -> None:
        """Add a LINE entity to the geometry layout. Removes parts of the line
        hidden by dimension text if `remove_hidden_lines` is True.

        Args:
            start: start point of line
            end: end point of line
            dxfattribs: additional or overridden DXF attributes
            remove_hidden_lines: removes parts of the line hidden by dimension
                text if ``True``

        """

        def add_line_to_block(start, end):
            # LINE is handled like an OCS entity !?
            self.layout.add_line(
                to_ocs(Vec3(start)).vec2,
                to_ocs(Vec3(end)).vec2,
                dxfattribs=dxfattribs,
            )

        def order(a: Vec2, b: Vec2) -> Tuple[Vec2, Vec2]:
            if (start - a).magnitude < (start - b).magnitude:
                return a, b
            else:
                return b, a

        to_ocs = self.ucs.to_ocs
        if remove_hidden_lines and self.has_text_box:
            text_box = self._text_box
            start_inside = int(text_box.is_inside(start))
            end_inside = int(text_box.is_inside(end))
            inside = start_inside + end_inside
            if inside == 2:  # start and end inside text_box
                return  # do not draw line
            elif inside == 1:  # one point inside text_box or on a border line
                intersection_points = text_box.intersect(
                    ConstructionLine(start, end)
                )
                if len(intersection_points) == 1:
                    # one point inside one point outside -> one intersection point
                    p1 = intersection_points[0]
                else:
                    # second point on a text box border line
                    p1, _ = order(*intersection_points)
                p2 = start if end_inside else end
                add_line_to_block(p1, p2)
                return
            else:
                intersection_points = text_box.intersect(
                    ConstructionLine(start, end)
                )
                if len(intersection_points) == 2:
                    # sort intersection points by distance to start point
                    p1, p2 = order(
                        intersection_points[0], intersection_points[1]
                    )
                    # line[start-p1] - gap - line[p2-end]
                    add_line_to_block(start, p1)
                    add_line_to_block(p2, end)
                    return
                # else: fall trough
        add_line_to_block(start, end)

    def add_arc(
        self,
        center: Vec2,
        radius: float,
        start_angle: float,
        end_angle: float,
        dxfattribs: dict = None,
        remove_hidden_lines=False,
    ) -> None:
        """Add a ARC entity to the geometry layout. Removes parts of the arc
        hidden by dimension text if `remove_hidden_lines` is True.

        Args:
            center: center of arc
            radius: radius of arc
            start_angle: start angle in radians
            end_angle: end angle in radians
            dxfattribs: additional or overridden DXF attributes
            remove_hidden_lines: removes parts of the arc hidden by dimension
                text if ``True``

        """

        def add_arc(s: float, e: float) -> None:
            """Add ARC entity to geometry block."""
            self.layout.add_arc(
                center=ocs_center,
                radius=radius,
                start_angle=math.degrees(ocs_angle(s)),
                end_angle=math.degrees(ocs_angle(e)),
                dxfattribs=dxfattribs,
            )

        # OCS of the ARC is defined by the DIMENSION entity!
        # Therefore remove OCS elevation, the elevation is defined by the
        # DIMENSION 'text_midpoint' (group code 11) and do not set 'extrusion'
        # either!
        ocs_center = self.ucs.to_ocs(Vec3(center)).vec2
        ocs_angle = self.ucs.to_ocs_angle_rad
        if remove_hidden_lines and self.has_text_box:
            for start, end in visible_arcs(
                center,
                radius,
                start_angle,
                end_angle,
                self._text_box,
            ):
                add_arc(start, end)
        else:
            add_arc(start_angle, end_angle)


class BaseDimensionRenderer:
    """Base rendering class for DIMENSION entities."""

    def __init__(
        self,
        dimension: Dimension,
        ucs: "UCS" = None,
        override: DimStyleOverride = None,
    ):
        self.dimension: Dimension = dimension
        self.geometry = self.init_geometry(dimension, ucs)

        # DimStyleOverride object, manages dimension style overriding
        self.dim_style: DimStyleOverride
        if override:
            self.dim_style = override
        else:
            self.dim_style = DimStyleOverride(dimension)

        # ---------------------------------------------
        # GENERAL PROPERTIES
        # ---------------------------------------------
        self.default_color: int = self.dimension.dxf.color  # ACI
        self.default_layer: str = self.dimension.dxf.layer

        get = self.dim_style.get
        # Overall scaling of DIMENSION entity:
        self.dim_scale: float = get("dimscale", 1.0)
        if abs(self.dim_scale) < 1e-9:
            self.dim_scale = 1.0

        # Controls drawing of circle or arc center marks and center lines, for
        # DIMDIAMETER and DIMRADIUS, the center mark is drawn only if you place
        # the dimension line outside the circle or arc.
        # 0 = No center marks or lines are drawn
        # <0 = Center lines are drawn
        # >0 = Center marks are drawn
        self.dim_center_marks: int = get("dimcen", 0)

        self.measurement = self.init_measurement(
            self.default_color, self.dim_scale
        )
        self.dimension_line: DimensionLine = self.init_dimension_line(
            self.default_color, self.dim_scale
        )
        self.arrows: Arrows = self.init_arrows(
            self.default_color, self.dim_scale
        )
        # Suppress arrow rendering - only rendering is suppressed (rendering
        # effect).
        # All placing related calculations are done without this settings.
        # Used for multi point linear dimensions to avoid double rendering of
        # non arrow ticks. These are ezdxf specific attributes!
        self.arrows.suppress1 = self.dim_style.pop("suppress_arrow1", False)
        self.arrows.suppress2 = self.dim_style.pop("suppress_arrow2", False)

        self.extension_lines: ExtensionLines = self.init_extension_lines(
            self.default_color, self.dim_scale
        )
        # tolerances have to be initialized after measurement:
        self.tol: Tolerance = self.init_tolerance(
            self.dim_scale, self.measurement
        )

        # Update text height
        self.measurement.text_height = max(
            self.measurement.text_height, self.tol.text_height
        )

    def init_geometry(self, dimension: Dimension, ucs: UCS = None):
        from ezdxf.layouts import VirtualLayout

        return Geometry(dimension, ucs or PassTroughUCS(), VirtualLayout())

    def init_tolerance(
        self, scale: float, measurement: Measurement
    ) -> Tolerance:
        return Tolerance(
            self.dim_style,
            cap_height=measurement.text_height,
            width_factor=measurement.text_width_factor,
            dim_scale=scale,
        )

    def init_extension_lines(self, color: int, scale: float) -> ExtensionLines:
        return ExtensionLines(self.dim_style, color, scale)

    def init_dimension_line(self, color: int, scale: float) -> DimensionLine:
        return DimensionLine(self.dim_style, color, scale)

    def init_arrows(self, color: int, scale: float) -> Arrows:
        return Arrows(self.dim_style, color, scale)

    def init_measurement(self, color: int, scale: float) -> Measurement:
        return LengthMeasurement(self.dim_style, color, scale)

    def init_text_box(self) -> TextBox:
        measurement = self.measurement
        return TextBox(
            center=measurement.text_location,
            width=self.total_text_width(),
            height=measurement.text_height,
            angle=measurement.text_rotation or 0.0,
            # The currently used mono-spaced abstract font, returns a too large
            # text width.
            # Therefore the horizontal text gap is ignored at all - yet!
            hgap=0.0,
            # Arbitrary choice to reduce the too large vertical gap!
            vgap=measurement.text_gap * 0.75,
        )

    def get_required_defpoint(self, name: str) -> Vec2:
        return get_required_defpoint(self.dimension, name)

    def render(self, block: "GenericLayoutType"):
        # Block entities are located in the OCS defined by the extrusion vector
        # of the DIMENSION entity and the z-axis of the OCS point
        # 'text_midpoint' (group code 11).
        self.geometry.set_layout(block)
        # Tolerance requires MTEXT support, switch off rendering of tolerances
        # and limits
        if not self.geometry.supports_dxf_r2000:
            self.tol.disable()

    def total_text_width(self) -> float:
        width = 0.0
        text = self.measurement.text
        if text:
            if self.tol.has_limits:  # only limits are displayed
                width = self.tol.text_width
            else:
                width = self.measurement.text_width(text)
                if self.tol.has_tolerance:
                    width += self.tol.text_width  # type: ignore
        return width

    def default_attributes(self) -> Dict[str, Any]:
        """Returns default DXF attributes as dict."""
        return {
            "layer": self.default_layer,
            "color": self.default_color,
        }

    def location_override(
        self, location: "Vertex", leader=False, relative=False
    ) -> None:
        """Set user defined dimension text location. ezdxf defines a user
        defined location per definition as 'outside'.

        Args:
            location: text midpoint
            leader: use leader or not (movement rules)
            relative: is location absolute (in UCS) or relative to dimension
                line center.

        """
        self.dim_style.set_location(location, leader, relative)
        self.measurement.location_override(location, leader, relative)

    def add_line(
        self,
        start: Vec2,
        end: Vec2,
        dxfattribs: Dict[str, Any],
        remove_hidden_lines=False,
    ) -> None:
        """Add a LINE entity to the dimension BLOCK. Removes parts of the line
        hidden by dimension text if `remove_hidden_lines` is True.

        Args:
            start: start point of line
            end: end point of line
            dxfattribs: additional or overridden DXF attributes
            remove_hidden_lines: removes parts of the line hidden by dimension
                text if ``True``

        """

        attribs = self.default_attributes()
        if dxfattribs:
            attribs.update(dxfattribs)
        self.geometry.add_line(start, end, dxfattribs, remove_hidden_lines)

    def add_arc(
        self,
        center: Vec2,
        radius: float,
        start_angle: float,
        end_angle: float,
        dxfattribs: Dict[str, Any] = None,
        remove_hidden_lines=False,
    ) -> None:
        """Add a ARC entity to the geometry layout. Removes parts of the arc
        hidden by dimension text if `remove_hidden_lines` is True.

        Args:
            center: center of arc
            radius: radius of arc
            start_angle: start angle in radians
            end_angle: end angle in radians
            dxfattribs: additional or overridden DXF attributes
            remove_hidden_lines: removes parts of the arc hidden by dimension
                text if ``True``

        """
        attribs = self.default_attributes()
        if dxfattribs:
            attribs.update(dxfattribs)
        self.geometry.add_arc(
            center, radius, start_angle, end_angle, attribs, remove_hidden_lines
        )

    def add_blockref(
        self,
        name: str,
        insert: Vec2,
        rotation: float,
        scale: float,
        dxfattribs: Dict[str, Any],
    ) -> None:
        """
        Add block references and standard arrows to the dimension BLOCK.

        Args:
            name: block or arrow name
            insert: insertion point in UCS
            rotation: rotation angle in degrees in UCS (x-axis is 0 degrees)
            scale: scaling factor for x- and y-direction
            dxfattribs: additional or overridden DXF attributes

        """
        attribs = self.default_attributes()
        if dxfattribs:
            attribs.update(dxfattribs)

        if name in ARROWS:
            # generates automatically BLOCK definitions for arrows if needed
            self.geometry.add_arrow_blockref(
                name, insert, scale, rotation, attribs
            )
        else:
            if name is None or not self.geometry.has_block(name):
                raise DXFUndefinedBlockError(f'Undefined block: "{name}"')
            if scale != 1.0:
                attribs["xscale"] = scale
                attribs["yscale"] = scale
            self.geometry.add_blockref(name, insert, rotation, attribs)

    def add_text(
        self, text: str, pos: Vec2, rotation: float, dxfattribs: Dict[str, Any]
    ) -> None:
        """
        Add TEXT (DXF R12) or MTEXT (DXF R2000+) entity to the dimension BLOCK.

        Args:
            text: text as string
            pos: insertion location in UCS
            rotation: rotation angle in degrees in UCS (x-axis is 0 degrees)
            dxfattribs: additional or overridden DXF attributes

        """
        geometry = self.geometry
        measurement = self.measurement
        attribs = self.default_attributes()
        attribs["style"] = measurement.text_style_name
        attribs["color"] = measurement.text_color

        if geometry.supports_dxf_r2000:  # use MTEXT entity
            attribs["char_height"] = measurement.text_height
            attribs["attachment_point"] = measurement.text_attachment_point
            if measurement.text_fill:
                attribs["box_fill_scale"] = measurement.text_box_fill_scale
                attribs["bg_fill_color"] = measurement.text_fill_color
                attribs["bg_fill"] = 3 if measurement.text_fill == 1 else 1

            if dxfattribs:
                attribs.update(dxfattribs)
            geometry.add_mtext(text, pos, rotation, dxfattribs=attribs)
        else:  # use TEXT entity
            attribs["height"] = measurement.text_height
            if dxfattribs:
                attribs.update(dxfattribs)
            geometry.add_text(text, pos, rotation, dxfattribs=attribs)

    def add_leader(self, p1: Vec2, p2: Vec2, p3: Vec2):
        """
        Add simple leader line from p1 to p2 to p3.

        Args:
            p1: target point
            p2: first text point
            p3: second text point

        """
        # use only  color and ignore linetype!
        dxfattribs = {"color": self.dimension_line.color}
        self.add_line(p1, p2, dxfattribs)
        self.add_line(p2, p3, dxfattribs)

    def transform_ucs_to_wcs(self) -> None:
        """Transforms dimension definition points into WCS or if required into
        OCS.

        Can not be called in __init__(), because inherited classes may be need
        unmodified values.

        """
        pass

    def finalize(self) -> None:
        self.transform_ucs_to_wcs()
        if self.geometry.requires_extrusion:
            self.dimension.dxf.extrusion = self.geometry.extrusion


def order_leader_points(p1: Vec2, p2: Vec2, p3: Vec2) -> Tuple[Vec2, Vec2]:
    if (p1 - p2).magnitude > (p1 - p3).magnitude:
        return p3, p2
    else:
        return p2, p3


def get_center_leader_points(
    target_point: Vec2, text_box: TextBox, leg_length: float
) -> Tuple[Vec2, Vec2]:
    """Returns the leader points of the "leg" for a vertical centered leader."""
    c0, c1, c2, c3 = text_box.corners
    #              c3-------c2
    # left leg /---x  text  x---\ right leg
    #         /    c0-------c1   \
    left_center = c0.lerp(c3)
    right_center = c1.lerp(c2)
    connection_point = left_center
    leg_vector = (c0 - c1).normalize(leg_length)
    if target_point.distance(left_center) > target_point.distance(right_center):
        connection_point = right_center
        leg_vector = -leg_vector
    # leader line: target_point -> leg_point -> connection_point
    # The text gap between the text and the connection point is already included
    # in the text_box corners!
    # Do not order leader points!
    return connection_point + leg_vector, connection_point


def get_required_defpoint(dim: Dimension, name: str) -> Vec2:
    dxf = dim.dxf
    if dxf.hasattr(name):  # has to exist, ignore default value!
        return Vec2(dxf.get(name))
    raise const.DXFMissingDefinitionPoint(name)


def visible_arcs(
    center: Vec2,
    radius: float,
    start_angle: float,
    end_angle: float,
    box: ConstructionBox,
) -> List[Tuple[float, float]]:
    """Returns the visible parts of an arc intersecting with a construction box
    as (start angle, end angle) tuples.

    Args:
        center: center of the arc
        radius: radius of the arc
        start_angle: start angle of arc in radians
        end_angle: end angle of arc in radians
        box: construction box which may intersect the arc

    """

    intersection_angles: List[float] = []  # angles are in the range 0 to 2pi
    start_angle %= math.tau
    end_angle %= math.tau
    arc = ConstructionArc(
        center, radius, math.degrees(start_angle), math.degrees(end_angle)
    )
    for line in box.border_lines():
        for intersection_point in arc.intersect_line(line):
            angle = (intersection_point - center).angle % math.tau
            if not intersection_angles:
                intersection_angles.append(angle)
            # new angle should be different than the last added angle:
            elif not math.isclose(intersection_angles[-1], angle):
                intersection_angles.append(angle)
    # Arc has to intersect the box in exact two locations!
    if len(intersection_angles) == 2:
        if start_angle > end_angle:  # arc passes 0 degrees
            intersection_angles = [
                (a if a >= start_angle else a + math.tau)
                for a in intersection_angles
            ]
        intersection_angles.sort()
        return [
            (start_angle, intersection_angles[0]),
            (intersection_angles[1], end_angle),
        ]
    else:
        # Ignore cases where the start- or the end point is inside the box.
        # Ignore cases where the box touches the arc in one point.
        return [(start_angle, end_angle)]


def get_text_style(doc: "Drawing", name: str) -> "Textstyle":
    assert doc is not None, "valid DXF document required"
    get_style = doc.tables.styles.get
    try:
        style = get_style(name)
    except const.DXFTableEntryError:
        style = get_style("Standard")
    return cast("Textstyle", style)


def get_char_height(
    dim_style: DimStyleOverride, text_style: "Textstyle"
) -> float:
    """Unscaled character height defined by text style or DIMTXT."""
    height: float = text_style.dxf.get("height", 0.0)
    if height == 0.0:  # variable text height (not fixed)
        height = dim_style.get("dimtxt", 1.0)
    return height


def compile_mtext(measurement: Measurement, tol: Tolerance) -> str:
    text = measurement.text
    if tol.enabled:
        text = tol.compile_mtext(text)
    return text



# Copyright (c) 2021, Manfred Moitzi
# License: MIT License
import math
from typing import TYPE_CHECKING, Tuple, List, Optional, Dict, Any
from abc import abstractmethod
import logging

from ezdxf.math import (
    Vec2,
    Vec3,
    NULLVEC,
    UCS,
    decdeg2dms,
    arc_angle_span_rad,
    xround,
)
from ezdxf.entities import DimStyleOverride, Dimension, DXFEntity
from .dim_base import (
    BaseDimensionRenderer,
    get_required_defpoint,
    format_text,
    apply_dimpost,
    TextBox,
    Tolerance,
    Measurement,
    LengthMeasurement,
    compile_mtext,
    order_leader_points,
    get_center_leader_points,
)
from ezdxf.render.arrows import ARROWS, arrow_length
from ezdxf.tools.text import is_upside_down_text_angle
from ezdxf.math import intersection_line_line_2d

if TYPE_CHECKING:
    from ezdxf.eztypes import GenericLayoutType

__all__ = ["AngularDimension", "Angular3PDimension", "ArcLengthDimension"]

logger = logging.getLogger("ezdxf")

ARC_PREFIX = "( "


def has_required_attributes(entity: DXFEntity, attrib_names: List[str]):
    has = entity.dxf.hasattr
    return all(has(attrib_name) for attrib_name in attrib_names)


GRAD = 200.0 / math.pi
DEG = 180.0 / math.pi


def format_angular_text(
    value: float,
    angle_units: int,
    dimrnd: Optional[float],
    dimdec: int,
    dimzin: int,
    dimdsep: str,
) -> str:
    def decimal_format(_value: float) -> str:
        return format_text(
            _value,
            dimrnd=dimrnd,
            dimdec=dimdec,
            dimzin=dimzin,
            dimdsep=dimdsep,
        )

    def dms_format(_value: float) -> str:
        if dimrnd is not None:
            _value = xround(_value, dimrnd)
        d, m, s = decdeg2dms(_value)
        if dimdec > 4:
            places = dimdec - 5
            s = round(s, places)
            return f"{d:.0f}°{m:.0f}'{decimal_format(s)}\""
        if dimdec > 2:
            return f"{d:.0f}°{m:.0f}'{s:.0f}\""
        if dimdec > 0:
            return f"{d:.0f}°{m:.0f}'"
        return f"{d:.0f}°"

    # angular_unit:
    # 0 = Decimal degrees
    # 1 = Degrees/minutes/seconds
    # 2 = Grad
    # 3 = Radians
    text = ""
    if angle_units == 0:
        text = decimal_format(value * DEG) + "°"
    elif angle_units == 1:
        text = dms_format(value * DEG)
    elif angle_units == 2:
        text = decimal_format(value * GRAD) + "g"
    elif angle_units == 3:
        text = decimal_format(value) + "r"
    return text


_ANGLE_UNITS = [
    DEG,
    DEG,
    GRAD,
    1.0,
]


def to_radians(value: float, dimaunit: int) -> float:
    try:
        return value / _ANGLE_UNITS[dimaunit]
    except IndexError:
        return value / DEG


class AngularTolerance(Tolerance):
    def __init__(
        self,
        dim_style: DimStyleOverride,
        cap_height: float = 1.0,
        width_factor: float = 1.0,
        dim_scale: float = 1.0,
        angle_units: int = 0,
    ):
        self.angular_units = angle_units
        super().__init__(dim_style, cap_height, width_factor, dim_scale)
        # Tolerance values are interpreted in dimaunit:
        # dimtp 1 means 1 degree for dimaunit = 0 or 1, but 1 radians for
        # dimaunit = 3
        # format_text() requires radians as input:
        self.update_tolerance_text(
            to_radians(self.maximum, angle_units),
            to_radians(self.minimum, angle_units),
        )

    def format_text(self, value: float) -> str:
        """Rounding and text formatting of tolerance `value`, removes leading
        and trailing zeros if necessary.

        """
        return format_angular_text(
            value=value,
            angle_units=self.angular_units,
            dimrnd=None,
            dimdec=self.decimal_places,
            dimzin=self.suppress_zeros,
            dimdsep=self.text_decimal_separator,
        )

    def update_limits(self, measurement: float) -> None:
        # measurement is in radians, tolerance values are interpreted in
        # dimaunit: dimtp 1 means 1 degree for dimaunit = 0 or 1,
        # but 1 radians for dimaunit = 3
        # format_text() requires radians as input:
        upper_limit = measurement + to_radians(self.maximum, self.angular_units)
        lower_limit = measurement - to_radians(self.minimum, self.angular_units)
        self.text_upper = self.format_text(upper_limit)
        self.text_lower = self.format_text(lower_limit)
        self.text_width = self.get_text_width(self.text_upper, self.text_lower)


class AngleMeasurement(Measurement):
    def update(self, raw_measurement_value: float) -> None:
        self.raw_value = raw_measurement_value
        self.value = raw_measurement_value
        self.text = self.text_override(raw_measurement_value)

    def format_text(self, value: float) -> str:
        text = format_angular_text(
            value=value,
            angle_units=self.angle_units,
            dimrnd=None,
            dimdec=self.angular_decimal_places,
            dimzin=self.angular_suppress_zeros << 2,  # convert to dimzin value
            dimdsep=self.decimal_separator,
        )
        if self.text_post_process_format:
            text = apply_dimpost(text, self.text_post_process_format)
        return text


def fits_into_arc_span(length: float, radius: float, arc_span: float) -> bool:
    required_arc_span: float = length / radius
    return arc_span > required_arc_span


class _CurvedDimensionLine(BaseDimensionRenderer):
    def __init__(
        self,
        dimension: Dimension,
        ucs: "UCS" = None,
        override: DimStyleOverride = None,
    ):
        super().__init__(dimension, ucs, override)
        # Common parameters for all sub-classes:
        # Use hidden line detection for dimension line:
        # Disable expensive hidden line calculation if possible!
        self.remove_hidden_lines_of_dimline = True
        self.center_of_arc: Vec2 = self.get_center_of_arc()
        self.dim_line_radius: float = self.get_dim_line_radius()
        self.ext1_dir: Vec2 = self.get_ext1_dir()
        self.start_angle_rad: float = self.ext1_dir.angle
        self.ext2_dir: Vec2 = self.get_ext2_dir()
        self.end_angle_rad: float = self.ext2_dir.angle

        # Angle between extension lines for all curved dimensions:
        # equal to the angle measurement of angular dimensions
        self.arc_angle_span_rad: float = arc_angle_span_rad(
            self.start_angle_rad, self.end_angle_rad
        )
        self.center_angle_rad = (
            self.start_angle_rad + self.arc_angle_span_rad / 2.0
        )

        # Additional required parameters but calculated later by sub-classes:
        self.ext1_start = Vec2()  # start of 1st extension line
        self.ext2_start = Vec2()  # start of 2nd extension line

        # Class specific setup:
        self.update_measurement()
        if self.tol.has_limits:
            self.tol.update_limits(self.measurement.value)

        # Text width and -height is required first, text location and -rotation
        # are not valid yet:
        self.text_box = self.init_text_box()

        # Place arrows outside?
        self.arrows_outside = False

        self.setup_text_and_arrow_fitting()
        self.setup_text_location()

        # update text box location and -rotation:
        self.text_box.center = self.measurement.text_location
        self.text_box.angle = self.measurement.text_rotation
        self.geometry.set_text_box(self.text_box)

        # Update final text location in the DIMENSION entity:
        self.dimension.dxf.text_midpoint = self.measurement.text_location

    @property
    def ocs_center_of_arc(self) -> Vec3:
        return self.geometry.ucs.to_ocs(Vec3(self.center_of_arc))

    @property
    def dim_midpoint(self) -> Vec2:
        """Return the midpoint of the dimension line."""
        return self.center_of_arc + Vec2.from_angle(
            self.center_angle_rad, self.dim_line_radius
        )

    @abstractmethod
    def update_measurement(self) -> None:
        """Setup measurement text."""
        ...

    @abstractmethod
    def get_ext1_dir(self) -> Vec2:
        """Return the direction of the 1st extension line == start angle."""
        ...

    @abstractmethod
    def get_ext2_dir(self) -> Vec2:
        """Return the direction of the 2nd extension line == end angle."""
        ...

    @abstractmethod
    def get_center_of_arc(self) -> Vec2:
        """Return the center of the arc."""
        ...

    @abstractmethod
    def get_dim_line_radius(self) -> float:
        """Return the distance from the center of the arc to the dimension line
        location
        """
        ...

    @abstractmethod
    def get_defpoints(self) -> List[Vec2]:
        ...

    def transform_ucs_to_wcs(self) -> None:
        """Transforms dimension definition points into WCS or if required into
        OCS.
        """

        def from_ucs(attr, func):
            if dxf.is_supported(attr):
                point = dxf.get(attr, NULLVEC)
                dxf.set(attr, func(point))

        dxf = self.dimension.dxf
        ucs = self.geometry.ucs
        from_ucs("defpoint", ucs.to_wcs)
        from_ucs("defpoint2", ucs.to_wcs)
        from_ucs("defpoint3", ucs.to_wcs)
        from_ucs("defpoint4", ucs.to_wcs)
        from_ucs("defpoint5", ucs.to_wcs)
        from_ucs("text_midpoint", ucs.to_ocs)

    def default_location(self, shift: float = 0.0) -> Vec2:
        radius = (
            self.dim_line_radius
            + self.measurement.text_vertical_distance()
            + shift
        )
        text_radial_dir = Vec2.from_angle(self.center_angle_rad)
        return self.center_of_arc + text_radial_dir * radius

    def setup_text_and_arrow_fitting(self):
        # self.text_box.width includes the gaps between text and dimension line
        # Is the measurement text without the arrows too wide to fit between the
        # extension lines?
        self.measurement.is_wide_text = not fits_into_arc_span(
            self.text_box.width, self.dim_line_radius, self.arc_angle_span_rad
        )

        required_text_and_arrows_space: float = (
            # The suppression of the arrows is not taken into account:
            self.text_box.width
            + 2.0 * self.arrows.arrow_size
        )

        # dimatfit: measurement text fitting rule is ignored!
        # Place arrows outside?
        self.arrows_outside: bool = not fits_into_arc_span(
            required_text_and_arrows_space,
            self.dim_line_radius,
            self.arc_angle_span_rad,
        )
        # Place measurement text outside?
        self.measurement.text_is_outside = not fits_into_arc_span(
            required_text_and_arrows_space * 1.1,  # add some extra space
            self.dim_line_radius,
            self.arc_angle_span_rad,
        )

        if (
            self.measurement.text_is_outside
            and self.measurement.user_text_rotation is None
        ):
            # Intersection of the measurement text with the dimension line is
            # not possible:
            self.remove_hidden_lines_of_dimline = False

    def setup_text_location(self) -> None:
        """Setup geometric text properties (location, rotation) and the TextBox
        object.
        """
        # dimtix: measurement.force_text_inside is ignored
        # dimtih: measurement.text_inside_horizontal is ignored
        # dimtoh: measurement.text_outside_horizontal is ignored

        # text radial direction = center -> text
        text_radial_dir: Vec2  # text "vertical" direction
        measurement = self.measurement

        # determine text location:
        at_default_location: bool = measurement.user_location is None
        has_text_shifting: bool = bool(
            measurement.text_shift_h or measurement.text_shift_v
        )
        if at_default_location:
            # place text in the "horizontal" center of the dimension line at the
            # default location defined by measurement.text_valign (dimtad):
            text_radial_dir = Vec2.from_angle(self.center_angle_rad)
            shift_text_upwards: float = 0.0
            if measurement.text_is_outside:
                # reset vertical alignment to "above"
                measurement.text_valign = 1
                if measurement.is_wide_text:
                    # move measurement text "above" the extension line endings:
                    shift_text_upwards = self.extension_lines.extension_above
            measurement.text_location = self.default_location(
                shift=shift_text_upwards
            )
            if (
                measurement.text_valign > 0 and not has_text_shifting
            ):  # not in the center and no text shifting is applied
                # disable expensive hidden line calculation
                self.remove_hidden_lines_of_dimline = False
        else:
            # apply dimtmove: measurement.text_movement_rule
            user_location = measurement.user_location
            assert isinstance(user_location, Vec2)
            if measurement.relative_user_location:
                user_location += self.dim_midpoint
            measurement.text_location = user_location
            if measurement.text_movement_rule == 0:
                # Moves the dimension line with dimension text and
                # aligns the text direction perpendicular to the connection
                # line from the arc center to the text center:
                self.dim_line_radius = (
                    self.center_of_arc - user_location
                ).magnitude
                # Attributes about the text and arrow fitting have to be
                # updated now:
                self.setup_text_and_arrow_fitting()
            elif measurement.text_movement_rule == 1:
                # Adds a leader when dimension text, text direction is
                # "horizontal" or user text rotation if given.
                # Leader location is defined by dimtad (text_valign):
                # "center" - connects to the left or right center of the text
                # "below" - add a line below the text
                if measurement.user_text_rotation is None:
                    # override text rotation
                    measurement.user_text_rotation = 0.0
                measurement.text_is_outside = True  # by definition
            elif measurement.text_movement_rule == 2:
                # Allows text to be moved freely without a leader and
                # aligns the text direction perpendicular to the connection
                # line from the arc center to the text center:
                measurement.text_is_outside = True  # by definition
            text_radial_dir = (
                measurement.text_location - self.center_of_arc
            ).normalize()

        # set text "horizontal":
        text_tangential_dir = text_radial_dir.orthogonal(ccw=False)

        if at_default_location and has_text_shifting:
            # Apply text relative shift (ezdxf only feature)
            if measurement.text_shift_h:
                measurement.text_location += (
                    text_tangential_dir * measurement.text_shift_h
                )
            if measurement.text_shift_v:
                measurement.text_location += (
                    text_radial_dir * measurement.text_shift_v
                )

        # apply user text rotation; rotation in degrees:
        if measurement.user_text_rotation is None:
            rotation = text_tangential_dir.angle_deg
        else:
            rotation = measurement.user_text_rotation

        if not self.geometry.requires_extrusion:
            # todo: extrusion vector (0, 0, -1)?
            # Practically all DIMENSION entities are 2D entities,
            # where OCS == WCS, check WCS text orientation:
            wcs_angle = self.geometry.ucs.to_ocs_angle_deg(rotation)
            if is_upside_down_text_angle(wcs_angle):
                measurement.has_upside_down_correction = True
                rotation += 180.0  # apply to UCS rotation!
        measurement.text_rotation = rotation

    def get_leader_points(self) -> Tuple[Vec2, Vec2]:
        # Leader location is defined by dimtad (text_valign):
        # "center":
        # - connects to the left or right vertical center of the text
        # - distance between text and leader line is measurement.text_gap (dimgap)
        #   and is already included in the text_box corner points
        # - length of "leg": arrows.arrow_size
        # "below" - add a line below the text
        if self.measurement.text_valign == 0:  # "center"
            return get_center_leader_points(
                self.dim_midpoint, self.text_box, self.arrows.arrow_size
            )
        else:  # "below"
            c0, c1, c2, c3 = self.text_box.corners
            if self.measurement.has_upside_down_correction:
                p1, p2 = c2, c3
            else:
                p1, p2 = c0, c1
            return order_leader_points(self.dim_midpoint, p1, p2)

    def render(self, block: "GenericLayoutType") -> None:
        """Main method to create dimension geometry of basic DXF entities in the
        associated BLOCK layout.

        Args:
            block: target BLOCK for rendering

        """
        super().render(block)
        self.add_extension_lines()
        adjust_start_angle, adjust_end_angle = self.add_arrows()

        measurement = self.measurement
        if measurement.text:
            if self.geometry.supports_dxf_r2000:
                text = compile_mtext(measurement, self.tol)
            else:
                text = measurement.text
            self.add_measurement_text(
                text, measurement.text_location, measurement.text_rotation
            )
            if measurement.has_leader:
                p1, p2 = self.get_leader_points()
                self.add_leader(self.dim_midpoint, p1, p2)
        self.add_dimension_line(adjust_start_angle, adjust_end_angle)
        self.geometry.add_defpoints(self.get_defpoints())

    def add_extension_lines(self) -> None:
        ext_lines = self.extension_lines
        if not ext_lines.suppress1:
            self._add_ext_line(
                self.ext1_start, self.ext1_dir, ext_lines.dxfattribs(1)
            )
        if not ext_lines.suppress2:
            self._add_ext_line(
                self.ext2_start, self.ext2_dir, ext_lines.dxfattribs(2)
            )

    def _add_ext_line(
        self, start: Vec2, direction: Vec2, dxfattribs: Dict[str, Any]
    ) -> None:
        ext_lines = self.extension_lines
        center = self.center_of_arc
        radius = self.dim_line_radius
        ext_above = ext_lines.extension_above
        is_inside = (start - center).magnitude > radius

        if ext_lines.has_fixed_length:
            ext_below = ext_lines.length_below
            if is_inside:
                ext_below, ext_above = ext_above, ext_below
            start = center + direction * (radius - ext_below)
        else:
            offset = ext_lines.offset
            if is_inside:
                ext_above = -ext_above
                offset = -offset
            start += direction * offset
        end = center + direction * (radius + ext_above)
        self.add_line(start, end, dxfattribs=dxfattribs)

    def add_arrows(self) -> Tuple[float, float]:
        """Add arrows or ticks to dimension.

        Returns: dimension start- and end angle offsets to adjust the
            dimension line

        """
        arrows = self.arrows
        attribs = arrows.dxfattribs()
        radius = self.dim_line_radius
        if abs(radius) < 1e-12:
            return 0.0, 0.0

        start = self.center_of_arc + self.ext1_dir * radius
        end = self.center_of_arc + self.ext2_dir * radius
        angle1 = self.ext1_dir.orthogonal().angle_deg
        angle2 = self.ext2_dir.orthogonal().angle_deg
        outside = self.arrows_outside
        arrow1 = not arrows.suppress1
        arrow2 = not arrows.suppress2
        start_angle_offset = 0.0
        end_angle_offset = 0.0
        if arrows.tick_size > 0.0:  # oblique stroke, but double the size
            if arrow1:
                self.add_blockref(
                    ARROWS.oblique,
                    insert=start,
                    rotation=angle1,
                    scale=arrows.tick_size * 2.0,
                    dxfattribs=attribs,
                )
            if arrow2:
                self.add_blockref(
                    ARROWS.oblique,
                    insert=end,
                    rotation=angle2,
                    scale=arrows.tick_size * 2.0,
                    dxfattribs=attribs,
                )
        else:
            arrow_size = arrows.arrow_size
            # Note: The arrow blocks are correct as they are!
            # The arrow head is tilted to match the connection point of the
            # dimension line (even for datum arrows).
            # tilting angle = 1/2 of the arc angle defined by the arrow length
            arrow_tilt: float = arrow_size / radius * 0.5 * DEG
            start_angle = angle1 + 180.0
            end_angle = angle2
            if outside:
                start_angle += 180.0
                end_angle += 180.0
                arrow_tilt = -arrow_tilt
            scale = arrow_size
            if arrow1:
                self.add_blockref(
                    arrows.arrow1_name,
                    insert=start,
                    scale=scale,
                    rotation=start_angle + arrow_tilt,
                    dxfattribs=attribs,
                )  # reverse
            if arrow2:
                self.add_blockref(
                    arrows.arrow2_name,
                    insert=end,
                    scale=scale,
                    rotation=end_angle - arrow_tilt,
                    dxfattribs=attribs,
                )
            if not outside:
                # arrows inside extension lines:
                # adjust angles for the remaining dimension line
                if arrow1:
                    start_angle_offset = (
                        arrow_length(arrows.arrow1_name, arrow_size) / radius
                    )
                if arrow2:
                    end_angle_offset = (
                        arrow_length(arrows.arrow2_name, arrow_size) / radius
                    )
        return start_angle_offset, end_angle_offset

    def add_dimension_line(
        self,
        start_offset: float,
        end_offset: float,
    ) -> None:
        # Start- and end angle adjustments have to be limited between the
        # extension lines.
        # Negative offset extends the dimension line outside!
        start_angle: float = self.start_angle_rad
        end_angle: float = self.end_angle_rad
        arrows = self.arrows
        size = arrows.arrow_size
        radius = self.dim_line_radius
        max_adjustment: float = abs(self.arc_angle_span_rad) / 2.0

        if start_offset > max_adjustment:
            start_offset = 0.0
        if end_offset > max_adjustment:
            end_offset = 0.0

        self.add_arc(
            self.center_of_arc,
            radius,
            start_angle + start_offset,
            end_angle - end_offset,
            dxfattribs=self.dimension_line.dxfattribs(),
            # hidden line detection if text is not placed outside:
            remove_hidden_lines=self.remove_hidden_lines_of_dimline,
        )
        if self.arrows_outside and not arrows.has_ticks:
            # add arrow extension lines
            start_offset, end_offset = arrow_offset_angles(
                arrows.arrow1_name, size, radius
            )
            self.add_arrow_extension_line(
                start_angle - end_offset,
                start_angle - start_offset,
            )
            start_offset, end_offset = arrow_offset_angles(
                arrows.arrow1_name, size, radius
            )
            self.add_arrow_extension_line(
                end_angle + start_offset,
                end_angle + end_offset,
            )

    def add_arrow_extension_line(self, start_angle: float, end_angle: float):
        self.add_arc(
            self.center_of_arc,
            self.dim_line_radius,
            start_angle=start_angle,
            end_angle=end_angle,
            dxfattribs=self.dimension_line.dxfattribs(),
        )

    def add_measurement_text(
        self, dim_text: str, pos: Vec2, rotation: float
    ) -> None:
        """Add measurement text to dimension BLOCK.

        Args:
            dim_text: dimension text
            pos: text location
            rotation: text rotation in degrees

        """
        attribs = self.measurement.dxfattribs()
        self.add_text(dim_text, pos=pos, rotation=rotation, dxfattribs=attribs)


class _AngularCommonBase(_CurvedDimensionLine):
    def init_tolerance(
        self, scale: float, measurement: Measurement
    ) -> Tolerance:
        return AngularTolerance(
            self.dim_style,
            cap_height=measurement.text_height,
            width_factor=measurement.text_width_factor,
            dim_scale=scale,
            angle_units=measurement.angle_units,
        )

    def init_measurement(self, color: int, scale: float) -> Measurement:
        return AngleMeasurement(
            self.dim_style, self.default_color, self.dim_scale
        )

    def update_measurement(self) -> None:
        self.measurement.update(self.arc_angle_span_rad)


class AngularDimension(_AngularCommonBase):
    """
    Angular dimension line renderer. The dimension line is defined by two lines.

    Supported render types:

    - default location above
    - default location center
    - user defined location, text aligned with dimension line
    - user defined location horizontal text

    Args:
        dimension: DIMENSION entity
        ucs: user defined coordinate system
        override: dimension style override management object

    """

    # Required defpoints:
    # defpoint = start point of 1st leg (group code 10)
    # defpoint4 = end point of 1st leg (group code 15)
    # defpoint3 = start point of 2nd leg (group code 14)
    # defpoint2 = end point of 2nd leg (group code 13)
    # defpoint5 = location of dimension line (group code 16)

    # unsupported or ignored features (at least by BricsCAD):
    # dimtih: text inside horizontal
    # dimtoh: text outside horizontal
    # dimjust: text position horizontal
    # dimdle: dimline extension

    def __init__(
        self,
        dimension: Dimension,
        ucs: "UCS" = None,
        override: DimStyleOverride = None,
    ):
        self.leg1_start = get_required_defpoint(dimension, "defpoint")
        self.leg1_end = get_required_defpoint(dimension, "defpoint4")
        self.leg2_start = get_required_defpoint(dimension, "defpoint3")
        self.leg2_end = get_required_defpoint(dimension, "defpoint2")
        self.dim_line_location = get_required_defpoint(dimension, "defpoint5")
        super().__init__(dimension, ucs, override)
        # The extension line parameters depending on the location of the
        # dimension line related to the definition point.
        # Detect the extension start point.
        # Which definition point is closer to the dimension line:
        self.ext1_start = detect_closer_defpoint(
            direction=self.ext1_dir,
            base=self.dim_line_location,
            p1=self.leg1_start,
            p2=self.leg1_end,
        )
        self.ext2_start = detect_closer_defpoint(
            direction=self.ext2_dir,
            base=self.dim_line_location,
            p1=self.leg2_start,
            p2=self.leg2_end,
        )

    def get_defpoints(self) -> List[Vec2]:
        return [
            self.leg1_start,
            self.leg1_end,
            self.leg2_start,
            self.leg2_end,
            self.dim_line_location,
        ]

    def get_center_of_arc(self) -> Vec2:
        center = intersection_line_line_2d(
            (self.leg1_start, self.leg1_end),
            (self.leg2_start, self.leg2_end),
        )
        if center is None:
            logger.warning(
                f"Invalid colinear or parallel angle legs found in {self.dimension})"
            )
            # This case can not be created by the GUI in BricsCAD, but DXF
            # files can contain any shit!
            # The interpolation of the end-points is an arbitrary choice and
            # maybe not the best choice!
            center = self.leg1_end.lerp(self.leg2_end)
        return center

    def get_dim_line_radius(self) -> float:
        return (self.dim_line_location - self.center_of_arc).magnitude

    def get_ext1_dir(self) -> Vec2:
        center = self.center_of_arc
        start = (
            self.leg1_end
            if self.leg1_start.isclose(center)
            else self.leg1_start
        )
        return (start - center).normalize()

    def get_ext2_dir(self) -> Vec2:
        center = self.center_of_arc
        start = (
            self.leg2_end
            if self.leg2_start.isclose(center)
            else self.leg2_start
        )
        return (start - center).normalize()


class Angular3PDimension(_AngularCommonBase):
    """
    Angular dimension line renderer. The dimension line is defined by three
    points.

    Supported render types:

    - default location above
    - default location center
    - user defined location, text aligned with dimension line
    - user defined location horizontal text

    Args:
        dimension: DIMENSION entity
        ucs: user defined coordinate system
        override: dimension style override management object

    """

    # Required defpoints:
    # defpoint = location of dimension line (group code 10)
    # defpoint2 = 1st leg (group code 13)
    # defpoint3 = 2nd leg (group code 14)
    # defpoint4 = center of angle (group code 15)

    def __init__(
        self,
        dimension: Dimension,
        ucs: "UCS" = None,
        override: DimStyleOverride = None,
    ):
        self.dim_line_location = get_required_defpoint(dimension, "defpoint")
        self.leg1_start = get_required_defpoint(dimension, "defpoint2")
        self.leg2_start = get_required_defpoint(dimension, "defpoint3")
        self.center_of_arc = get_required_defpoint(dimension, "defpoint4")
        super().__init__(dimension, ucs, override)
        self.ext1_start = self.leg1_start
        self.ext2_start = self.leg2_start

    def get_defpoints(self) -> List[Vec2]:
        return [
            self.dim_line_location,
            self.leg1_start,
            self.leg2_start,
            self.center_of_arc,
        ]

    def get_center_of_arc(self) -> Vec2:
        return self.center_of_arc

    def get_dim_line_radius(self) -> float:
        return (self.dim_line_location - self.center_of_arc).magnitude

    def get_ext1_dir(self) -> Vec2:
        return (self.leg1_start - self.center_of_arc).normalize()

    def get_ext2_dir(self) -> Vec2:
        return (self.leg2_start - self.center_of_arc).normalize()


class ArcLengthMeasurement(LengthMeasurement):
    def format_text(self, value: float) -> str:
        text = format_text(
            value=value,
            dimrnd=self.text_round,
            dimdec=self.decimal_places,
            dimzin=self.suppress_zeros,
            dimdsep=self.decimal_separator,
        )
        if self.has_arc_length_prefix:
            text = ARC_PREFIX + text
        if self.text_post_process_format:
            text = apply_dimpost(text, self.text_post_process_format)
        return text


class ArcLengthDimension(_CurvedDimensionLine):
    """
    Arc length dimension line renderer.
    Requires DXF R2004.

    Supported render types:

    - default location above
    - default location center
    - user defined location, text aligned with dimension line
    - user defined location horizontal text

    Args:
        dimension: DXF entity DIMENSION
        ucs: user defined coordinate system
        override: dimension style override management object

    """

    # Required defpoints:
    # defpoint = location of dimension line (group code 10)
    # defpoint2 = 1st arc point (group code 13)
    # defpoint3 = 2nd arc point (group code 14)
    # defpoint4 = center of arc (group code 15)

    def __init__(
        self,
        dimension: Dimension,
        ucs: "UCS" = None,
        override: DimStyleOverride = None,
    ):
        self.dim_line_location = get_required_defpoint(dimension, "defpoint")
        self.leg1_start = get_required_defpoint(dimension, "defpoint2")
        self.leg2_start = get_required_defpoint(dimension, "defpoint3")
        self.center_of_arc = get_required_defpoint(dimension, "defpoint4")
        self.arc_radius = (self.leg1_start - self.center_of_arc).magnitude
        super().__init__(dimension, ucs, override)
        self.ext1_start = self.leg1_start
        self.ext2_start = self.leg2_start

    def get_defpoints(self) -> List[Vec2]:
        return [
            self.dim_line_location,
            self.leg1_start,
            self.leg2_start,
            self.center_of_arc,
        ]

    def init_measurement(self, color: int, scale: float) -> Measurement:
        return ArcLengthMeasurement(
            self.dim_style, self.default_color, self.dim_scale
        )

    def get_center_of_arc(self) -> Vec2:
        return self.center_of_arc

    def get_dim_line_radius(self) -> float:
        return (self.dim_line_location - self.center_of_arc).magnitude

    def get_ext1_dir(self) -> Vec2:
        return (self.leg1_start - self.center_of_arc).normalize()

    def get_ext2_dir(self) -> Vec2:
        return (self.leg2_start - self.center_of_arc).normalize()

    def update_measurement(self) -> None:
        angle = arc_angle_span_rad(self.start_angle_rad, self.end_angle_rad)
        arc_length = angle * self.arc_radius
        self.measurement.update(arc_length)


def detect_closer_defpoint(
    direction: Vec2, base: Vec2, p1: Vec2, p2: Vec2
) -> Vec2:
    # Calculate the projected distance onto the (normalized) direction vector:
    d0 = direction.dot(base)
    d1 = direction.dot(p1)
    d2 = direction.dot(p2)
    # Which defpoint is closer to the base point (d0)?
    if abs(d1 - d0) <= abs(d2 - d0):
        return p1
    return p2


def arrow_offset_angles(
    arrow_name: str, size: float, radius: float
) -> Tuple[float, float]:
    start_offset: float = 0.0
    end_offset: float = size / radius
    length = arrow_length(arrow_name, size)
    if length > 0.0:
        start_offset = length / radius
        end_offset *= 2.0
    return start_offset, end_offset



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import Dict, NamedTuple, Optional, List


class Bookmark(NamedTuple):
    name: str
    handle: str
    offset: int


class Bookmarks:
    def __init__(self):
        self.bookmarks: Dict[str, Bookmark] = dict()

    def add(self, name: str, handle: str, offset: int):
        self.bookmarks[name] = Bookmark(name, handle, offset)

    def get(self, name: str) -> Optional[Bookmark]:
        return self.bookmarks.get(name)

    def names(self) -> List[str]:
        return list(self.bookmarks.keys())

    def discard(self, name: str):
        try:
            del self.bookmarks[name]
        except KeyError:
            pass

    def clear(self):
        self.bookmarks.clear()



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
"""
Block Reference Management
==========================

The package `ezdxf` is not designed as a CAD library and does not automatically
monitor all internal changes. This enables faster entity processing at the cost
of an unknown state of the DXF document.

In order to carry out precise BLOCK reference management, i.e. to handle
dependencies or to delete unused BLOCK definition, the block reference status
(counter) must be acquired explicitly by the package user.
All block reference management structures must be explicitly recreated each time
the document content is changed. This is not very efficient, but it is safe.

.. warning::

    And even with all this careful approach, it is always possible to destroy a
    DXF document by deleting an absolutely necessary block definition.

Always remember that `ezdxf` is not intended or suitable as a basis for a CAD
application!

.. versionadded:: 0.18

"""
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Iterator
from collections import Counter

from ezdxf.lldxf.types import POINTER_CODES
from ezdxf.protocols import referenced_blocks

if TYPE_CHECKING:
    from ezdxf.eztypes import Drawing, Tags, DXFEntity, BlockRecord

__all__ = ["BlockDefinitionIndex", "BlockReferenceCounter"]

""" 
Where are block references located:

- HEADER SECTION: $DIMBLK, $DIMBLK1, $DIMBLK2, $DIMLDRBLK 
- DIMENSION: arrows referenced in the associated anonymous BLOCK, covered by 
  the INSERT entities in that BLOCK
- ACAD_TABLE: has an anonymous BLOCK representation, covered by the 
  INSERT entities in that BLOCK
- LEADER: DIMSTYLE override "dimldrblk" is stored as handle in XDATA

Entity specific block references, returned by the "ReferencedBlocks" protocol:
- INSERT: "name"
- DIMSTYLE: "dimblk", "dimblk1", "dimblk2", "dimldrblk"
- MLEADER: arrows, blocks - has no anonymous BLOCK representation
- MLEADERSTYLE: arrows
- DIMENSION: "geometry", the associated anonymous BLOCK
- ACAD_TABLE: currently managed as generic DXFTagStorage, that should return 
  the group code 343 "block_record" 

Possible unknown or undocumented block references:
- DXFTagStorage - all handle group codes
- XDATA - only group code 1005
- APPDATA - all handle group codes
- XRECORD - all handle group codes

Contains no block references as far as known:
- DICTIONARY: can only references DXF objects like XRECORD or DICTIONARYVAR
- Extension Dictionary is a DICTIONARY object
- REACTORS - used for object messaging, a reactor does not establish 
  a block reference

Block references are stored as handles to the BLOCK_RECORD entity!

Testing DXF documents with missing BLOCK definitions:

- INSERT without an existing BLOCK definition does NOT crash AutoCAD/BricsCAD
- HEADER variables $DIMBLK, $DIMBLK2, $DIMBLK2 and $DIMLDRBLK can reference 
  non existing blocks without crashing AutoCAD/BricsCAD  

"""


class BlockDefinitionIndex:
    """Index of all :class:`~ezdxf.entities.BlockRecord` entities representing
    real BLOCK definitions, excluding all :class:`~ezdxf.entities.BlockRecord`
    entities defining model space or paper space layouts. External references
    (XREF) and XREF overlays are included.

    """
    def __init__(self, doc: "Drawing"):
        self._doc = doc
        # mapping: handle -> BlockRecord entity
        self._handle_index: Dict[str, "BlockRecord"] = dict()
        # mapping: block name -> BlockRecord entity
        self._name_index: Dict[str, "BlockRecord"] = dict()
        self.rebuild()

    @property
    def block_records(self) -> Iterator["BlockRecord"]:
        """Returns an iterator of all :class:`~ezdxf.entities.BlockRecord`
        entities representing BLOCK definitions.
        """
        return iter(self._doc.tables.block_records)  # type: ignore

    def rebuild(self):
        """Rebuild index from scratch."""
        handle_index = self._handle_index
        name_index = self._name_index
        handle_index.clear()
        name_index.clear()
        for block_record in self.block_records:
            if block_record.is_block_layout:
                handle_index[block_record.dxf.handle] = block_record
                name_index[block_record.dxf.name] = block_record

    def has_handle(self, handle: str) -> bool:
        """Returns ``True`` if a :class:`~ezdxf.entities.BlockRecord` for the
        given block record handle exist.
        """
        return handle in self._handle_index

    def has_name(self, name: str) -> bool:
        """Returns ``True`` if a :class:`~ezdxf.entities.BlockRecord` for the
        given block name exist.
        """
        return name in self._name_index

    def by_handle(self, handle: str) -> Optional["BlockRecord"]:
        """Returns the :class:`~ezdxf.entities.BlockRecord` for the given block
        record handle or ``None``.
        """
        return self._handle_index.get(handle)

    def by_name(self, name: str) -> Optional["BlockRecord"]:
        """Returns :class:`~ezdxf.entities.BlockRecord` for the given block name
        or ``None``.
        """
        return self._name_index.get(name)


class BlockReferenceCounter:
    """
    Counts all block references in a DXF document.

    Check if a block is referenced by any entity or any resource (DIMSYTLE,
    MLEADERSTYLE) in a DXF document::

        import ezdxf
        from ezdxf.blkrefs import BlockReferenceCounter

        doc = ezdxf.readfile("your.dxf")
        counter = BlockReferenceCounter(doc)
        count = counter.by_name("XYZ")
        print(f"Block 'XYZ' if referenced {count} times.")

    """

    def __init__(self, doc: "Drawing", index: BlockDefinitionIndex = None):
        # mapping: handle -> BlockRecord entity
        self._block_record_index = (
            index if index is not None else BlockDefinitionIndex(doc)
        )

        # mapping: handle -> reference count
        self._counter = count_references(
            doc.entitydb.values(), self._block_record_index
        )
        self._counter.update(header_section_handles(doc))

    def by_handle(self, handle: str) -> int:
        """Returns the block reference count for a given
        :class:`~ezdxf.entities.BlockRecord` handle.
        """
        return self._counter[handle]

    def by_name(self, block_name: str) -> int:
        """Returns the block reference count for a given block name."""
        handle = ""
        block_record = self._block_record_index.by_name(block_name)
        if block_record is not None:
            handle = block_record.dxf.handle
        return self._counter[handle]


def count_references(
    entities: Iterable["DXFEntity"], index: BlockDefinitionIndex
) -> Counter:
    from ezdxf.entities import XRecord, DXFTagStorage

    def update(handles: Iterable[str]):
        # only count references to existing blocks:
        counter.update(h for h in handles if index.has_handle(h))

    counter: Counter = Counter()
    for entity in entities:
        # add handles stored in XDATA and APP data
        update(generic_handles(entity))
        # add entity specific block references
        update(referenced_blocks(entity))
        # special entity types storing arbitrary raw DXF tags:
        if isinstance(entity, XRecord):
            update(all_pointer_handles(entity.tags))
        elif isinstance(entity, DXFTagStorage):
            # XDATA and APP data is already done!
            for tags in entity.xtags.subclasses[1:]:
                update(all_pointer_handles(tags))
            # ignore embedded objects: special objects for MTEXT and ATTRIB
    return counter


def generic_handles(entity: "DXFEntity") -> Iterable[str]:
    handles: List[str] = []
    if entity.xdata is not None:
        for tags in entity.xdata.data.values():
            handles.extend(value for code, value in tags if code == 1005)
    if entity.appdata is not None:
        for tags in entity.appdata.data.values():
            handles.extend(all_pointer_handles(tags))
    return handles


def all_pointer_handles(tags: "Tags") -> Iterable[str]:
    return (value for code, value in tags if code in POINTER_CODES)


def header_section_handles(doc: "Drawing") -> Iterable[str]:
    header = doc.header
    for var_name in ("$DIMBLK", "$DIMBLK1", "$DIMBLK2", "$DIMLDRBLK"):
        blk_name = header.get(var_name, None)
        if blk_name is not None:
            block = doc.blocks.get(blk_name, None)
            if block is not None:
                yield block.block_record.dxf.handle



# Purpose: Grouping entities by DXF attributes or a key function.
# Copyright (c) 2017-2021, Manfred Moitzi
# License: MIT License
from typing import Iterable, Hashable, Dict, List, TYPE_CHECKING

from ezdxf.lldxf.const import DXFValueError, DXFAttributeError

if TYPE_CHECKING:
    from ezdxf.eztypes import DXFEntity, KeyFunc


def groupby(
    entities: Iterable["DXFEntity"], dxfattrib: str = "", key: "KeyFunc" = None
) -> Dict[Hashable, List["DXFEntity"]]:
    """
    Groups a sequence of DXF entities by a DXF attribute like ``'layer'``,
    returns a dict with `dxfattrib` values as key and a list of entities
    matching this `dxfattrib`.
    A `key` function can be used to combine some DXF attributes (e.g. layer and
    color) and should return a hashable data type like a tuple of strings,
    integers or floats, `key` function example::

        def group_key(entity: DXFEntity):
            return entity.dxf.layer, entity.dxf.color

    For not suitable DXF entities return ``None`` to exclude this entity, in
    this case it's not required, because :func:`groupby` catches
    :class:`DXFAttributeError` exceptions to exclude entities, which do not
    provide layer and/or color attributes, automatically.

    Result dict for `dxfattrib` = ``'layer'`` may look like this::

        {
            '0': [ ... list of entities ],
            'ExampleLayer1': [ ... ],
            'ExampleLayer2': [ ... ],
            ...
        }

    Result dict for `key` = `group_key`, which returns a ``(layer, color)``
    tuple, may look like this::

        {
            ('0', 1): [ ... list of entities ],
            ('0', 3): [ ... ],
            ('0', 7): [ ... ],
            ('ExampleLayer1', 1): [ ... ],
            ('ExampleLayer1', 2): [ ... ],
            ('ExampleLayer1', 5): [ ... ],
            ('ExampleLayer2', 7): [ ... ],
            ...
        }

    All entity containers (modelspace, paperspace layouts and blocks) and the
    :class:`~ezdxf.query.EntityQuery` object have a dedicated :meth:`groupby`
    method.

    Args:
        entities: sequence of DXF entities to group by a DXF attribute or a
            `key` function
        dxfattrib: grouping DXF attribute like ``'layer'``
        key: key function, which accepts a :class:`DXFEntity` as argument and
            returns a hashable grouping key or ``None`` to ignore this entity

    """
    if all((dxfattrib, key)):
        raise DXFValueError(
            "Specify a dxfattrib or a key function, but not both."
        )
    if dxfattrib != "":
        key = lambda entity: entity.dxf.get_default(dxfattrib)
    if key is None:
        raise DXFValueError(
            "no valid argument found, specify a dxfattrib or a key function, "
            "but not both."
        )

    result: Dict[Hashable, List["DXFEntity"]] = dict()
    for dxf_entity in entities:
        if not dxf_entity.is_alive:
            continue
        try:
            group_key = key(dxf_entity)
        except DXFAttributeError:
            # ignore DXF entities, which do not support all query attributes
            continue
        if group_key is not None:
            group = result.setdefault(group_key, [])
            group.append(dxf_entity)
    return result



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

from typing import TYPE_CHECKING, List
import logging
import math
from ezdxf.math import Vec2, UCS, NULLVEC
from ezdxf.lldxf import const
from ezdxf.entities import DimStyleOverride, Dimension
from .dim_base import (
    BaseDimensionRenderer,
    get_required_defpoint,
    compile_mtext,
)

if TYPE_CHECKING:
    from ezdxf.eztypes import GenericLayoutType

__all__ = ["OrdinateDimension"]

logger = logging.getLogger("ezdxf")


class OrdinateDimension(BaseDimensionRenderer):
    # Required defpoints:
    # defpoint = origin (group code 10)
    # defpoint2 = feature location (group code 13)
    # defpoint3 = end of leader (group code 14)
    # user text location is ignored (group code 11) and replaced by default
    # location calculated by the ezdxf renderer:

    def __init__(
        self,
        dimension: Dimension,
        ucs: "UCS" = None,
        override: DimStyleOverride = None,
    ):
        # The local coordinate system is defined by origin and the
        # horizontal_direction in OCS:
        self.origin_ocs: Vec2 = get_required_defpoint(dimension, "defpoint")
        self.feature_location_ocs: Vec2 = get_required_defpoint(
            dimension, "defpoint2"
        )
        self.end_of_leader_ocs: Vec2 = get_required_defpoint(
            dimension, "defpoint3"
        )
        # Horizontal direction in clockwise orientation, see DXF reference
        # for group code 51:
        self.horizontal_dir = -dimension.dxf.get("horizontal_direction", 0.0)
        self.rotation = math.radians(self.horizontal_dir)
        self.local_x_axis = Vec2.from_angle(self.rotation)
        self.local_y_axis = self.local_x_axis.orthogonal()
        self.x_type = bool(  # x-type is set!
            dimension.dxf.get("dimtype", 0) & const.DIM_ORDINATE_TYPE
        )
        super().__init__(dimension, ucs, override)

        # Measurement directions can be opposite to local x- or y-axis
        self.leader_vec_ocs = self.end_of_leader_ocs - self.feature_location_ocs
        leader_x_vec = self.local_x_axis.project(self.leader_vec_ocs)
        leader_y_vec = self.local_y_axis.project(self.leader_vec_ocs)
        try:
            self.measurement_direction: Vec2 = leader_x_vec.normalize()
        except ZeroDivisionError:
            self.measurement_direction = Vec2(1, 0)
        try:
            self.measurement_orthogonal: Vec2 = leader_y_vec.normalize()
        except ZeroDivisionError:
            self.measurement_orthogonal = Vec2(0, 1)

        if not self.x_type:
            self.measurement_direction, self.measurement_orthogonal = (
                self.measurement_orthogonal,
                self.measurement_direction,
            )

        self.update_measurement()
        if self.tol.has_limits:
            self.tol.update_limits(self.measurement.value)

        # Text width and -height is required first, text location and -rotation
        # are not valid yet:
        self.text_box = self.init_text_box()

        # Set text location and rotation:
        self.measurement.text_location = self.get_default_text_location()
        self.measurement.text_rotation = self.get_default_text_rotation()

        # Update text box location and -rotation:
        self.text_box.center = self.measurement.text_location
        self.text_box.angle = self.measurement.text_rotation
        self.geometry.set_text_box(self.text_box)

        # Update final text location in the DIMENSION entity:
        self.dimension.dxf.text_midpoint = self.measurement.text_location

    def get_default_text_location(self) -> Vec2:
        if self.x_type:
            text_vertical_shifting_dir = -self.local_x_axis
        else:
            text_vertical_shifting_dir = self.local_y_axis

        # user text location is not supported and ignored:
        return (
            self.end_of_leader_ocs
            + self.measurement_orthogonal * (self.text_box.width * 0.5)
            + text_vertical_shifting_dir
            * self.measurement.text_vertical_distance()
        )

    def get_default_text_rotation(self) -> float:
        # user text rotation is not supported and ignored:
        return (90.0 if self.x_type else 0.0) + self.horizontal_dir

    def update_measurement(self) -> None:
        feature_location_vec: Vec2 = self.feature_location_ocs - self.origin_ocs
        # ordinate measurement is always absolute:
        self.measurement.update(
            self.local_x_axis.project(feature_location_vec).magnitude
            if self.x_type
            else self.local_y_axis.project(feature_location_vec).magnitude
        )

    def get_defpoints(self) -> List[Vec2]:
        return [
            self.origin_ocs,
            self.feature_location_ocs,
            self.end_of_leader_ocs,
        ]

    def transform_ucs_to_wcs(self) -> None:
        """Transforms dimension definition points into WCS or if required into
        OCS.
        """

        def from_ucs(attr, func):
            point = dxf.get(attr, NULLVEC)
            dxf.set(attr, func(point))

        dxf = self.dimension.dxf
        ucs = self.geometry.ucs
        from_ucs("defpoint", ucs.to_wcs)
        from_ucs("defpoint2", ucs.to_wcs)
        from_ucs("defpoint3", ucs.to_wcs)
        from_ucs("text_midpoint", ucs.to_ocs)

        # Horizontal direction in clockwise orientation, see DXF reference
        # for group code 51:
        dxf.horizontal_direction = -ucs.to_ocs_angle_deg(self.horizontal_dir)

    def render(self, block: "GenericLayoutType") -> None:
        """Main method to create dimension geometry of basic DXF entities in the
        associated BLOCK layout.

        Args:
            block: target BLOCK for rendering

        """
        super().render(block)
        self.add_ordinate_leader()
        measurement = self.measurement
        if measurement.text:
            if self.geometry.supports_dxf_r2000:
                text = compile_mtext(measurement, self.tol)
            else:
                text = measurement.text
            self.add_measurement_text(
                text, measurement.text_location, measurement.text_rotation
            )
        self.geometry.add_defpoints(self.get_defpoints())

    def add_ordinate_leader(self) -> None:
        # DXF attributes from first extension line not from dimension line!
        attribs = self.extension_lines.dxfattribs(1)
        # The ordinate leader is normal to the measurement direction.
        # leader direction and text direction:
        direction = self.measurement_orthogonal
        leg_size = self.arrows.arrow_size * 2.0
        #            /---1---TEXT
        # x----0----/
        # d0 = distance from feature location (x) to 1st upward junction
        d0 = direction.project(self.leader_vec_ocs).magnitude - 2.0 * leg_size

        start0 = (
            self.feature_location_ocs + direction * self.extension_lines.offset
        )
        end0 = self.feature_location_ocs + direction * max(leg_size, d0)
        start1 = self.end_of_leader_ocs - direction * leg_size
        end1 = self.end_of_leader_ocs
        if self.measurement.vertical_placement != 0:
            end1 += direction * self.text_box.width

        self.add_line(start0, end0, dxfattribs=attribs)
        self.add_line(end0, start1, dxfattribs=attribs)
        self.add_line(start1, end1, dxfattribs=attribs)

    def add_measurement_text(
        self, dim_text: str, pos: Vec2, rotation: float
    ) -> None:
        """Add measurement text to dimension BLOCK.

        Args:
            dim_text: dimension text
            pos: text location
            rotation: text rotation in degrees

        """
        attribs = self.measurement.dxfattribs()
        self.add_text(dim_text, pos=pos, rotation=rotation, dxfattribs=attribs)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

"""
Resource management module for transferring DXF resources between documents.

Planning state!!!

"""
from typing import Dict, List, Iterator, cast, Callable, Iterable
from ezdxf.respkg import RTP
from ezdxf.entities import DXFEntity, factory
from ezdxf.document import Drawing

# The goal is that this module should not be imported from any module which
# implements DXF entities, so that this module can import necessary resource
# entities like Layer, Linetype or MLeaderStyle

__all__ = ["ResourceTransferManager"]

"""
Extension dictionaries: handled in DXFEntity.copy() and factory.bind()

Reactors? 
I think reactors in resources can be dropped. 
Under no circumstances should reactors initiate a copy process of the 
referenced entities.

Block references? 
The BLOCK_RECORD is a table entry and somehow a resource. 
But the BLOCK_RECORD also represents paperspace layouts, block layouts, the
modelspace and whole documents as XREFs.

Handles in the XDATA section? 
Handle references in XDATA in order of appearance:

class DXFEntity:
    def dump_resources(self, package: ResourcePackage):
        package.push(self.get_entities_in_xdata())

    def get_entities_in_xdata(self):
        # in order of appearance
        e1 = 1. 1005 -> 1. get entity by handle
        e2 = 2. 1003 -> 2. get layer entity by name
        e3 = 3. 1005 -> 3. get entity by handle
        return [e1, e2, e3]
    
    def put_entities_in_xdata(self, entities):
        for e in entities:
            # 1. 1005 <- 1. entity.dxf.handle
            # 2. 1003 <- 2. layer.dxf.name
            # 3. 1005 <- 3. entity.dxf.handle

    def load_resources(self, package: ResourcePackage):
        xdata = package.pop()
        self.put_entities_in_xdata(xdata)  

Each subclass puts its own list on top, variable data has to be appended at 
the end and has to be checked by isinstance().
  
class Layer(DXFEntity):
    def dump_resources(self, package: ResourcePackage):
        super().dump_resources(package)
        linetype = self.doc.linetypes.get(self.dxf.linetype)
        package.push([linetype])

    def load_resources(self, package: ResourcePackage):
        linetype, *_ = package.pop()
        self.dxf.name = linetype.dxf.name
        super().load_resources(package)

"""


class _ResourcePackage:
    def __init__(self, entity: DXFEntity):
        self.entity = entity
        self._resources: List[List[DXFEntity]] = list()

    def push(self, resources: List[DXFEntity]):
        self._resources.append(resources)

    def pop(self) -> List[DXFEntity]:
        return self._resources.pop()

    def resources(self) -> Iterator[DXFEntity]:
        for data in self._resources:
            yield from data

    def foreach(self, func: Callable[[DXFEntity], DXFEntity]):
        self._resources = [
            [func(entity) for entity in entities]
            for entities in self._resources
        ]


class ResourceTransferManager:
    def __init__(self):
        self.packages: Dict[int, _ResourcePackage] = dict()
        self.copy_machine = CopyMachine()

    def register(self, entity: DXFEntity):
        """Register resources."""
        if id(entity) not in self.packages:
            package = _ResourcePackage(entity)
            assert isinstance(entity, RTP)
            entity.dump_resources(package)
            self.packages[id(entity)] = package

    def transfer(self, target: Drawing):
        """Transfer resources if the transaction can be finished completely.
        """
        self.copy_all()
        if self.copy_machine.has_finished():
            # TODO: order of execution?
            self.swap_resources()
            self.update_entities()
            bind(self.copy_machine.copies, target)
            assign_resources(self.copy_machine.copies, target)

    def copy_all(self):
        """Copy resources."""
        for pkg in self.packages.values():
            # Important: copy only the resources, the entity copy itself is
            # another task!
            self.copy_machine.append(pkg.resources())
        self.copy_machine.run()

    def swap_resources(self):
        """Replace resources by copies."""
        for pkg in self.packages.values():
            pkg.foreach(self.copy_machine.swap)

    def update_entities(self):
        """Update entity resources."""
        for pkg in self.packages.values():
            rtp = cast("RTP", pkg.entity)
            rtp.load_resources(pkg)


class CopyMachine:
    def __init__(self):
        self._copies: Dict[int, DXFEntity] = dict()
        self._originals: Dict[int, DXFEntity] = dict()

    @property
    def copies(self) -> Iterable[DXFEntity]:
        """Returns the copies"""
        return self._copies.values()

    def has_finished(self) -> bool:
        """All resources are copied?"""
        return not len(self._originals)

    def append(self, entities: Iterable[DXFEntity]):
        """Append entities to copy."""
        for e in entities:
            if id(e) not in self._copies:
                self._originals[id(e)] = e

    def run(self):
        """Execute copy process."""
        trash = list()
        new = list()
        while self._originals:
            # do not add/delete while iterating
            trash.clear()
            new.clear()
            for entity in self._originals.values():
                pkg = _ResourcePackage(entity)
                entity.dump_resources(pkg)  # type: ignore
                resources = list(pkg.resources())
                if all(id(e) in self._copies for e in resources):
                    # all resources are copied
                    trash.append(self.copy_entity(entity))
                else:
                    # add required resources to copy
                    new.extend(resources)
            count = len(self._originals)
            stuck = True
            self.append(new)
            if len(self._originals) != count:
                stuck = False
            for _id in trash:
                if _id:
                    del self._originals[_id]
            if stuck and len(self._originals) == count:
                break

    def copy_entity(self, entity: DXFEntity) -> int:
        """Copy a single entity and returns the id or 0 if the entity is
        already copied.
        """
        if id(entity) in self._copies:
            return 0
        self._copies[id(entity)] = entity.copy()
        return id(entity)

    def swap(self, entity: DXFEntity) -> DXFEntity:
        """Swap original entity against copy."""
        return self._copies[id(entity)]


def bind(entities: Iterable[DXFEntity], doc: Drawing):
    """Bind all entities to the DXF document."""
    for entity in entities:
        factory.bind(entity, doc)


def assign_resources(entities: Iterable[DXFEntity], doc: Drawing):
    """Add entities to the correct resource tables."""
    pass
